from collections import defaultdict
import itertools
import logging
import os
import pkgutil
import re
import sqlite3
import sys
from urllib.parse import quote as urlquote

from Bio.Seq import Seq
from Bio.SeqFeature import CompoundLocation
from Bio.SeqFeature import FeatureLocation
import pandas as pd

# COMPATIBILITY
SUPPORTED_DB_VERSION = 4


class sonarDBManager:
    """
    This object provides a sonarDB SQLite manager handler managing connections and
    providing context-safe transaction control.

    Notes
    -----
            This object should be called using a context manager to ensure rollbacks
            on abnormal termination.

    Example
    -------

    In this example the DOCTESTDB variable store the path to a database file

    >>> with sonarDBManager(DOCTESTDB) as dbm:
    ... 	pass

    Parameters
    ----------

    dbfile : str
            define a path to a non-existent or valid SONAR database file. If the
            file does not exist, a SONAR database is created.
    timeout : int [ -1 ]
            define busy timeout. Use -1 for no timeout.
    readonly : bool [ False ]
            define if the connection should be read-only
    debug : bool [ False ]
            debug mode (print selected sql queries)

    Attributes
    ----------
    dbfile : str
            stores the path to the used SONAR database file.
    connection : object
            stores the SQLite3 connection
    cursor : method
            stores the SQLite3 cursor

    Dev Note
    --------
    A database row is returned as dictionary with column name as keys. Multiple
    rows are returned as list of those dictionaries.

    """

    def __init__(
        self, dbfile, timeout=-1, readonly=True, debug=False, autocreate=False
    ):
        logging.basicConfig(format="%(asctime)s %(message)s")
        self.connection = None
        if not autocreate and not os.path.isfile(dbfile):
            sys.exit("database error: database does not exists")
        self.dbfile = os.path.abspath(dbfile)
        self.cursor = None
        self.__timeout = timeout
        self.__mode = "ro" if readonly else "rwc"
        self.__uri = self.get_uri(dbfile)
        self.debug = debug
        self.__properties = False
        self.__illegal_properties = {}
        self.__default_reference_id = False
        self.__lineage_sublineage_dict = None
        self.__rootdir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.lineagewithsublineages = os.path.join(
            self.__rootdir, "lib", "lineage.all.tsv"
        )

        self.operators = {
            "genuine": {
                "=": "=",
                ">": ">",
                "<": "<",
                ">=": ">=",
                "<=": "<=",
                "IN": "IN",
                "LIKE": "LIKE",
                "BETWEEN": "BETWEEN",
            },
            "negated": {
                "=": "!=",
                ">": "<=",
                "<": ">=",
                ">=": "<",
                "<=": ">",
                "IN": "NOT IN",
                "LIKE": "NOT LIKE",
                "BETWEEN": "NOT BETWEEN",
            },
        }

    def __enter__(self):
        self.connection, self.cursor = self.connect()
        self.start_transaction()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if [exc_type, exc_value, exc_traceback].count(None) != 3:
            if self.__mode == "rwc":
                print("rollback database", file=sys.stderr)
                self.rollback()
        elif self.__mode == "rwc":
            self.commit()
        self.close()

    def __del__(self):
        if self.connection:
            self.close()

    def connect(self):
        con = sqlite3.connect(
            self.__uri + "?mode=" + self.__mode,
            self.__timeout,
            isolation_level=None,
            uri=True,
        )
        if self.debug:
            con.set_trace_callback(logging.warning)
        con.row_factory = self.dict_factory
        cur = con.cursor()
        return con, cur

    def start_transaction(self):
        self.cursor.execute("BEGIN DEFERRED")

    def new_transaction(self):
        self.commit()
        self.start_transaction()

    def commit(self):
        self.connection.commit()

    def rollback(self):
        self.connection.rollback()

    def close(self):
        self.connection.close()

    def get_db_version(self):
        return self.cursor.execute("pragma user_version").fetchone()["user_version"]

    def check_db_compatibility(self):
        dbver = self.get_db_version()
        if dbver != SUPPORTED_DB_VERSION:
            sys.exit(
                "compatibility error: the given database is not compatible with this version of sonar (database version: "
                + str(dbver)
                + "; supported database version: "
                + str(SUPPORTED_DB_VERSION)
                + ")"
            )

    @property
    def lineage_sublineage_dict(self):
        if not self.__lineage_sublineage_dict:
            df = pd.read_sql(
                "SELECT * FROM lineages", self.connection
            )  # pd.read_csv(self.lineagewithsublineages, sep='\t')
            self.__lineage_sublineage_dict = dict(zip(df.lineage, df.sublineage))
        return self.__lineage_sublineage_dict

    @property
    def default_reference_id(self):
        if self.__default_reference is False:
            sql = "SELECT accession FROM reference WHERE standard = 1 LIMIT 1"
            self.__default_reference = self.cursor.execute(sql).fetchone()
        return self.__default_reference

    @property
    def properties(self):
        if self.__properties is False:
            sql = "SELECT * FROM property"
            rows = self.cursor.execute(sql).fetchall()
            self.__properties = {} if not rows else {x["name"]: x for x in rows}
        return self.__properties

    # SETUP

    @staticmethod
    def get_uri(dbfile):
        return "file:" + urlquote(dbfile)

    @staticmethod
    def setup(filename, debug=False):
        sql = pkgutil.get_data(__name__, "data/db.sqlite").decode()
        uri = sonarDBManager.get_uri(filename)
        with sqlite3.connect(uri + "?mode=rwc", uri=True) as con:
            if debug:
                con.set_trace_callback(logging.warning)
            con.executescript(sql)

    def add_codon(self, translation_table, codon, aa):
        sql = "INSERT OR IGNORE INTO translation (id, codon, aa) VALUES(?, ?, ?);"
        self.cursor.execute(sql, [translation_table, codon, aa])

    def add_property(self, name, datatype, querytype, description, standard=None):
        name = name.upper()
        if name in self.__illegal_properties:
            sys.exit(
                "error: '"
                + str(name)
                + "' is reserved and cannot be used as property name"
            )
        if not re.match("^[a-zA-Z0-9_]+$", name):
            sys.exit(
                "error: invalid property name (property names can contain only letters, numbers and underscores)"
            )
        if name in self.properties:
            sys.exit(
                "error: a property named "
                + name
                + " already exists in the given database."
            )
        try:
            sql = "INSERT INTO property (name, datatype, querytype, description, standard) VALUES(?, ?, ?, ?, ?);"
            self.cursor.execute(sql, [name, datatype, querytype, description, standard])
            self.__properties = False
            pid = self.properties[name]["id"]
            if standard is not None:
                sql = (
                    "INSERT INTO sample2property (property_id, value_"
                    + self.properties[name]["datatype"]
                    + ", sample_id) SELECT ?, ?, id FROM sample WHERE 1"
                )
                vals = [pid, standard]
                self.cursor.execute(sql, vals)
        except sqlite3.Error as error:
            print("Failed to insert data into sqlite table", error)
        return pid

    def add_translation_table(self, translation_table):
        sql = "SELECT COUNT(*) FROM translation  WHERE id = ?;"
        if self.cursor.execute(sql, [translation_table]).fetchone()["COUNT(*)"] != 4096:
            for codon in itertools.product("ATGCRYSWKMBDHVN-", repeat=3):
                codon = "".join(codon)
                try:
                    aa = str(Seq.translate(codon, table=translation_table))
                except Exception:
                    aa = ""
                self.add_codon(translation_table, codon, aa)

    def add_reference(
        self, accession, description, organism, translation_table, standard=0
    ):
        self.add_translation_table(translation_table)
        if standard:
            sql = "UPDATE reference SET standard = 0 WHERE standard != 0"
            self.cursor.execute(sql)
        sql = "INSERT INTO reference (id, accession, description, organism, translation_id, standard) VALUES(?, ?, ?, ?, ?, ?);"
        self.cursor.execute(
            sql, [None, accession, description, organism, translation_table, standard]
        )
        return self.cursor.lastrowid

    # INSERTING DATA

    def insert_property(self, sample_id, property_name, property_value):
        sql = (
            "INSERT OR REPLACE INTO sample2property (sample_id, property_id, value_"
            + self.properties[property_name]["datatype"]
            + ") VALUES(?, ?, ?);"
        )
        # print([sample_id, self.properties[property_name]['id'], property_value])
        self.cursor.execute(
            sql, [sample_id, self.properties[property_name]["id"], property_value]
        )
        return

    def insert_sequence(self, seqhash):
        sql = "INSERT OR IGNORE INTO sequence (seqhash) VALUES(?);"
        self.cursor.execute(sql, [seqhash])
        return seqhash

    def insert_sample(self, sample_name, seqhash):
        self.insert_sequence(seqhash)
        sql = "INSERT OR REPLACE INTO sample (name, seqhash, datahash) VALUES(?, ?, ?);"
        self.cursor.execute(sql, [sample_name, seqhash, ""])
        sql = "SELECT id FROM sample WHERE name = ?;"
        sid = self.cursor.execute(sql, [sample_name]).fetchone()["id"]
        for pname in self.properties:
            if not self.properties[pname]["standard"] is None:
                self.insert_property(sid, pname, self.properties[pname]["standard"])
        return

    def insert_alignment(self, seqhash, element_id):
        sql = (
            "INSERT OR IGNORE INTO alignment (id, seqhash, element_id) VALUES(?, ?, ?);"
        )
        self.cursor.execute(sql, [None, seqhash, element_id])
        sql = "SELECT id FROM alignment WHERE element_id = ? AND seqhash = ?;"
        return self.cursor.execute(sql, [element_id, seqhash]).fetchone()["id"]

    def insert_molecule(
        self,
        reference_id,
        type,
        accession,
        symbol,
        description,
        segment,
        length,
        standard=0,
    ):
        if symbol.strip() == "":
            symbol = accession
        if standard:
            sql = "UPDATE molecule SET standard = ? WHERE reference_id = ? AND standard = 1"
            self.cursor.execute(sql, [0, reference_id])
        sql = "INSERT INTO molecule (id, reference_id, type, accession, symbol, description, segment, length, standard) VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?);"
        self.cursor.execute(
            sql,
            [
                None,
                reference_id,
                type,
                accession,
                symbol,
                description,
                segment,
                length,
                standard,
            ],
        )
        sql = "SELECT id FROM molecule WHERE accession = ?"
        mid = self.cursor.execute(sql, [accession]).fetchone()["id"]
        return mid

    def insert_element(
        self,
        molecule_id,
        type,
        accession,
        symbol,
        description,
        start,
        end,
        strand,
        sequence,
        standard=0,
        parent_id=" ",
        parts=None,
    ):
        if symbol.strip() == "":
            symbol = accession
        if standard:
            sql = (
                "UPDATE element SET standard = ? WHERE molecule_id = ? AND standard = 1"
            )
            self.cursor.execute(sql, [0, molecule_id])
        sql = "INSERT INTO element (id, molecule_id, type, accession, symbol, description, start, end, strand, sequence, standard, parent_id) VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);"
        self.cursor.execute(
            sql,
            [
                None,
                molecule_id,
                type,
                accession,
                symbol,
                description,
                start,
                end,
                strand,
                sequence,
                standard,
                parent_id,
            ],
        )
        sql = "SELECT id FROM element WHERE accession = ?"
        mid = self.cursor.execute(sql, [accession]).fetchone()["id"]
        if parts is not None:
            for part in parts:
                sql = "INSERT OR IGNORE INTO elempart (element_id, start, end, strand, base, segment) VALUES(?, ?, ?, ?, ?, ?);"
                self.cursor.execute(sql, [mid] + part)
        return mid

    def insert_variant(
        self, alignment_id, element_id, ref, alt, start, end, label, parent_id=""
    ):
        sql = "INSERT OR IGNORE INTO variant (id, element_id, start, end, ref, alt, label, parent_id) VALUES(?, ?, ?, ?, ?, ?, ?, ?);"
        self.cursor.execute(
            sql, [None, element_id, start, end, ref, alt, label, parent_id]
        )
        vid = self.get_variant_id(element_id, start, end, ref, alt)
        sql = "INSERT OR IGNORE INTO alignment2variant (alignment_id, variant_id) VALUES(?, ?);"
        self.cursor.execute(sql, [alignment_id, vid])
        return vid

    # DELETING DATA

    def delete_samples(self, *sample_names):
        sql = (
            "DELETE FROM sample WHERE name IN ("
            + ", ".join(["?"] * len(sample_names))
            + ");"
        )
        self.cursor.execute(sql, sample_names)

    def delete_property(self, property_name):
        sql = "DELETE FROM sample2property WHERE property_id = ?;"
        self.cursor.execute(sql, [self.properties[property_name]["id"]])
        sql = "DELETE FROM property WHERE name = ?;"
        self.cursor.execute(sql, [property_name])

    # SELECTING DATA

    def sample_exists(self, sample_name):
        sql = "SELECT EXISTS(SELECT 1 FROM sample WHERE name=? LIMIT 1) as found"
        return bool(self.cursor.execute(sql, [sample_name]).fetchone()["found"])

    def get_sample_id(self, sample_name):
        sql = "SELECT id FROM sample WHERE name = ? LIMIT 1;"
        row = self.cursor.execute(sql, [sample_name]).fetchone()
        return row["id"] if row else None

    def get_sample_data(self, sample_name):
        sql = "SELECT id, seqhash FROM sample WHERE name = ? LIMIT 1;"
        row = self.cursor.execute(sql, [sample_name]).fetchone()
        return (row["id"], row["seqhash"]) if row else (None, None)

    def seq_exists(self, seqhash):
        sql = "SELECT id FROM sample WHERE seqhash=? LIMIT 1;"
        return False if self.cursor.execute(sql, [seqhash]).fetchone() is None else True

    def get_alignment_id(self, seqhash, element_id):
        sql = "SELECT id FROM alignment WHERE element_id = ? AND seqhash = ? LIMIT 1;"
        row = self.cursor.execute(sql, [element_id, seqhash]).fetchone()
        return None if row is None else row["id"]

    def get_default_reference_accession(self):
        sql = "SELECT accession FROM reference WHERE standard=1"
        return self.cursor.execute(sql).fetchone()["accession"]

    def get_molecule_ids(self, reference_accession=None):
        if reference_accession:
            condition = '"reference.accession" = ?'
            val = [reference_accession]
        else:
            condition = '"reference.standard" = ?'
            val = [1]
        sql = (
            'SELECT "molecule.accession", "molecule.id" FROM referenceView WHERE '
            + condition
        )
        return {
            x["molecule.accession"]: x["molecule.id"]
            for x in self.cursor.execute(sql, val).fetchall()
            if x is not None
        }

    def get_molecule_id(self, molecule_accession):
        sql = "SELECT id FROM molecule WHERE accession = ?;"
        row = self.cursor.execute(sql, [molecule_accession]).fetchone()
        if row:
            row = row["id"]
        return row

    def get_molecule_data(self, *fields, reference_accession=None):
        if not fields:
            fields = "*"
        elif '"molecule.accession"' not in fields:
            fields = list(fields) + ['"molecule.accession"']
        if reference_accession:
            condition = '"reference.accession" = ?'
            vals = [reference_accession]
        else:
            condition = '"reference.standard" = ?'
            vals = [1]
        sql = (
            "SELECT "
            + ", ".join(fields)
            + " FROM referenceView WHERE "
            + condition
            + ";"
        )
        row = self.cursor.execute(sql, vals).fetchall()
        if row:
            return {x["molecule.accession"]: x for x in row}
        return {}

    def get_elements(self, molecule_id, *types):
        sql = "SELECT * FROM element WHERE molecule_id = ?"
        if types:
            sql += " AND type IN (" + ", ".join(["?"] * len(types)) + ");"
        row = self.cursor.execute(sql, [molecule_id] + list(types)).fetchall()
        if not row:
            return []
        return row

    def get_element_ids(self, reference_accession=None, type=None):
        molecule_ids = list(
            self.get_molecule_ids(reference_accession=reference_accession).values()
        )
        sql = (
            'SELECT id FROM element WHERE "molecule_id" IN ('
            + ", ".join(["?"] * len(molecule_ids))
            + ")"
        )
        if type:
            sql += " AND type = ?"
            molecule_ids.append(type)
        row = self.cursor.execute(sql, molecule_ids).fetchall()
        if not row:
            return []
        return [x["id"] for x in row]

    def get_source(self, molecule_id):
        return self.get_elements(molecule_id, "source")[0]

    def get_annotation(
        self,
        reference_accession=None,
        molecule_accession=None,
        element_accession=None,
        element_type=None,
        fields=["*"],
    ):
        conditions = []
        vals = []
        if reference_accession:
            conditions.append('"reference.accession" = ?')
            vals.append(reference_accession)
        else:
            conditions.append('"reference.standard" = ?')
            vals.append(1)
        if molecule_accession:
            conditions.append('"molecule.accession" = ?')
            vals.append(molecule_accession)
        else:
            conditions.append('"molecule.standard" = ?')
            vals.append(1)
        if element_accession:
            conditions.append('"element.accession" = ?')
            vals.append(element_accession)
        elif not element_type:
            conditions.append('"element.type" = ?')
            vals.append("source")
        if element_type:
            conditions.append('"element.type" = ?')
            vals.append(element_type)
        sql = (
            "SELECT "
            + ", ".join(fields)
            + " FROM referenceView WHERE "
            + " AND ".join(conditions)
            + ' ORDER BY "reference.id" ASC, "molecule.id" ASC, "element.id" ASC, "element.segment" ASC'
        )
        return self.cursor.execute(sql, vals).fetchall()

    def get_alignment_data(self, sample_name, reference_accession=None):
        if reference_accession is None:
            sql = 'SELECT "reference.accession" as acc FROM alignmentView WHERE "sample.name" = ? LIMIT 1'
            reference_accession = self.cursor.execute(sql, [sample_name]).fetchone()
            if reference_accession is None:
                return ""
            reference_accession = reference_accession["acc"]
        sql = 'SELECT "element.sequence", "element.symbol", "element.id" FROM alignmentView WHERE "sample.name" = ? AND "reference.accession" = ? '
        return self.cursor.execute(sql, [sample_name, reference_accession])

    def get_variant_id(self, element_id, start, end, ref, alt):
        sql = "SELECT id FROM variant WHERE element_id = ? AND start = ? AND end = ? AND ref = ? AND alt = ?;"
        row = self.cursor.execute(sql, [element_id, start, end, ref, alt]).fetchone()
        return None if row is None else row["id"]

    def iter_dna_variants(self, sample_name, *element_ids):
        if len(element_ids) == 1:
            condition = " = ?"
        elif len(element_ids) > 1:
            condition = " IN (" + ", ".join(["?"] * len(element_ids)) + ")"
        else:
            condition = ""
        # sql = "SELECT \"element.id\", \"variant.start\", \"variant.end\", \"variant.ref\", \"variant.alt\" FROM variantView WHERE \"sample.name\" = ? AND \"element.id\"" + condition
        sql = (
            """
            SELECT  variant.element_id as \"element.id\",
                    variant.start as \"variant.start\",
                    variant.end as  \"variant.end\",
                    variant.ref as  \"variant.ref\",
                    variant.alt as \"variant.alt\"
                    FROM
                        ( SELECT sample.seqhash
                        FROM sample
                        WHERE sample.name = ?
                        ) AS sample_T
                    INNER JOIN alignment
                        ON sample_T.seqhash == alignment.seqhash
                    INNER JOIN alignment2variant
                        ON alignment.id == alignment2variant.alignment_id
                    INNER JOIN    variant
                        ON alignment2variant.variant_id == variant.id
                        WHERE  variant.element_id """
            + condition
        )

        for row in self.cursor.execute(sql, [sample_name] + list(element_ids)):
            if row["variant.start"] is not None:
                yield row

    def iter_profile(self, sample_name, element_id):
        sql = 'SELECT * FROM variantView WHERE "sample.name" = ? AND "element.id" = ? ORDER BY "element.id", "variant.start";'
        for row in self.cursor.execute(sql, [sample_name, element_id]):
            yield row

    def iter_protein_variants(self, sample_name, *element_ids):
        sql = (
            "SELECT element_id, element.symbol, start, end, ref, alt FROM variant WHERE name = ? AND element.id IN ("
            + ", ".join(["?"] * len(element_ids))
            + ") AND type = 'cds' LEFT JOIN element ON element.id = element_id;"
        )
        for row in self.cursor.execute(sql, [sample_name] + element_ids):
            yield row

    def get_frameshift_vars(self, sample_name, *element_ids):
        sql = (
            "SELECT element_id, element.symbol, start, end, ref, alt FROM variant WHERE name = ? AND element.id IN ("
            + ", ".join(["?"] * len(element_ids))
            + ") AND frameshift = 1 LEFT JOIN element ON element.id = element_id;"
        )
        for row in self.cursor.execute(sql, [sample_name] + element_ids):
            yield row

    def count_samples(self):
        sql = "SELECT COUNT(*) FROM sample;"
        return self.cursor.execute(sql).fetchone()["COUNT(*)"]

    def count_sequences(self):
        sql = "SELECT COUNT(DISTINCT seqhash) FROM sample;"
        return self.cursor.execute(sql).fetchone()["COUNT(DISTINCT seqhash)"]

    def count_property(self, property_name, distinct=False, ignore_standard=False):
        d = "DISTINCT " if distinct else ""
        c = "WHERE property_id = ?"
        v = [self.properties[property_name]["id"]]
        if ignore_standard and self.properties[property_name]["standard"] is not None:
            c += " AND value_" + self.properties[property_name]["datatype"] + " != ?"
            v.append(self.properties[property_name]["standard"])
        sql = (
            "SELECT COUNT("
            + d
            + " value_"
            + self.properties[property_name]["datatype"]
            + ") as count FROM sample2property "
            + c
            + ";"
        )
        return self.cursor.execute(sql, v).fetchone()["count"]

    def get_properties(self, sample_name, property_name=None):
        if property_name:
            sql = 'SELECT * FROM propertyView WHERE "sample.name" = ? AND property_name = ?;'
            vals = [sample_name, property_name]
        else:
            sql = 'SELECT * FROM propertyView WHERE "sample.name" = ?'
            vals = [sample_name]
        properties = {}
        for row in self.cursor.execute(sql, vals):
            if row is None:
                return {}
            properties[row["property.name"]] = row[
                "sample2property.value_" + row["property.datatype"]
            ]
        return properties

    def count_properties(self):
        sql = "SELECT COUNT(*) as count FROM propertyView GROUP BY property_name;"
        return self.cursor.execute(sql).fetchall()

    def get_seqhash(self, sample_name):
        sql = 'SELECT "sample.seqhash" FROM sequenceView WHERE sample.name = ?'
        return [
            x["sample.hash"]
            for x in self.cursor.execute(sql, [sample_name]).fetchall()
            if x is not None
        ]

    def get_translation_dict(self, translation_id):
        sql = "SELECT codon, aa FROM translation WHERE id = ?"
        return {
            x["codon"]: x["aa"]
            for x in self.cursor.execute(sql, [translation_id]).fetchall()
        }

    def get_earliest_import(self):
        sql = "SELECT MIN(imported) as import FROM genome WHERE import IS NOT NULL;"
        return self.cursor.execute(sql).fetchone()["import"]

    def get_latest_import(self):
        sql = "SELECT MAX(imported) as import FROM genome WHERE import IS NOT NULL;"
        return self.cursor.execute(sql).fetchone()["import"]

    def get_earliest_date(self):
        sql = "SELECT MIN(date) as date FROM genome WHERE date IS NOT NULL;"
        return self.cursor.execute(sql).fetchone()["date"]

    def get_latest_date(self):
        sql = "SELECT MAX(date) as date FROM genome WHERE date IS NOT NULL;"
        return self.cursor.execute(sql).fetchone()["date"]

    def get_element_parts(self, element_id=None):
        sql = "SELECT start, end, strand FROM elempart WHERE element_id = ? ORDER BY segment"
        return self.cursor.execute(sql, [element_id]).fetchall()

    def get_sequence(self, element_id=None):
        sql = "SELECT sequence, type FROM element WHERE id = ?"
        row = self.cursor.execute(sql, [element_id]).fetchone()
        return None if row is None else row["sequence"]

    def extract_sequence(self, element_id=None, translation_table=None):
        sql = "SELECT sequence, type, id, parent_id FROM element WHERE id = ?"
        row = self.cursor.execute(sql, [element_id]).fetchone()
        if not row:
            return None
        element_id = row["id"]
        while row and row["type"] not in {"source", "CDS"}:
            sql = "SELECT sequence, type, parent_id FROM element WHERE id = ?"
            row = self.cursor.execute(sql, [row["parent_id"]]).fetchone()
        sequence = row["sequence"]
        parts = []
        for part in self.get_element_parts(element_id):
            parts.append(
                FeatureLocation(part["start"], part["end"], strand=part["strand"])
            )
        feat = CompoundLocation(parts) if len(parts) > 1 else parts[0]
        if translation_table is None:
            return str(feat.extract(sequence))
        return str(
            Seq(feat.extract(sequence)).translate(
                table=translation_table, stop_symbol=""
            )
        )

    # Add/Update lineages into Table
    def add_update_lineage(self, _df):
        print("Prepare:", len(_df))

        # _df.to_sql(name='lineages', con = self.connection, if_exists='replace',
        # 			 index=False)
        sql = "INSERT OR REPLACE INTO lineages (lineage, sublineage) VALUES(?, ?);"
        for ind in _df.index:
            self.cursor.execute(sql, (_df["lineage"][ind], _df["sublineage"][ind]))
        # self.commit()
        # sql = "INSERT OR REPLACE INTO sample2property (sample_id, property_id, value_" + self.properties[property_name]['datatype'] + ") VALUES(?, ?, ?);"
        # print([sample_id, self.properties[property_name]['id'], property_value])
        # self.cursor.execute(sql, [sample_id, self.properties[property_name]['id'], property_value])

    # MATCHING PROFILES
    def get_operator(self, val):
        if val.startswith("^"):
            return val[1:], self.operators["negated"]
        return val, self.operators["genuine"]

    def get_conditional_expr(self, field, operator, *vals):
        condlist = []
        vallist = []

        # transforming operators
        if operator == "=" and len(vals) > 1:
            operator = "IN"
        elif operator == "!=" and len(vals) > 1:
            operator = "NOT IN"

        # creating conditions
        if operator == "IN" or operator == "NOT IN":
            condlist.append(
                field + " " + operator + "(" + ", ".join(["?"] * len(vals)) + ")"
            )
            vallist.extend(vals)

        elif operator == "BETWEEN" or operator == "NOT BETWEEN":
            for val in vals:
                condlist.append(field + " " + operator + " ? AND ?")
                vallist.extend(val)

        else:
            condlist.extend([field + " " + operator + " ?"] * len(vals))
            vallist.extend(vals)

        return condlist, vallist

    def query_numeric(self, field, *vals, link="OR"):
        link = " " + link.strip() + " "
        defaultop = "="
        op1 = re.compile(r"^(\^*)((?:>|>=|<|<=|!=|=)?)(-?[1-9]+[0-9]*)$")
        op2 = re.compile(r"^(\^*)(-?[1-9]+[0-9]*):(-?[1-9]+[0-9]*)$")
        errmsg = (
            "query error: numeric value or range expected for field " + field + "(got: "
        )
        data = defaultdict(set)
        conditions = []
        vallist = []

        for val in vals:
            val = str(val).strip()

            # single value
            if ":" not in val:
                match = op1.match(val)
                if not match:
                    sys.exit(errmsg + val + ")")
                op = match.group(2) if match.group(2) else defaultop
                operator = (
                    self.operators["negated"][op]
                    if match.group(1)
                    else self.operators["genuine"][op]
                )
                val = match.group(3)
                data[operator].add(val)

            # range
            else:
                match = op2.match(val)
                if not match:
                    sys.exit(errmsg + val + ")")
                operator = (
                    self.operators["negated"]["BETWEEN"]
                    if match.group(1)
                    else self.operators["genuine"]["BETWEEN"]
                )
                val = (match.group(2), match.group(3))
                data[operator].add(val)

        for operator, valset in data.items():
            condition, vals = self.get_conditional_expr(field, operator, *valset)
            conditions.extend(condition)
            vallist.extend(vals)

        return link.join(conditions), vallist

    def query_float(self, field, *vals, link="OR"):
        link = " " + link.strip() + " "
        defaultop = "="
        op1 = re.compile(r"^(\^*)((?:>|>=|<|<=|!=|=)?)(-?[1-9]+[0-9]*(?:.[0-9]+)*)$")
        op2 = re.compile(
            r"^(\^*)(-?[1-9]+[0-9]*(?:.[0-9]+)*):(-?[1-9]+[0-9]*(?:.[0-9]+)*)$"
        )
        errmsg = (
            "query error: decimal value or range expected for field " + field + "(got: "
        )
        data = defaultdict(set)
        conditions = []
        vallist = []

        for val in vals:
            val = str(val).strip()
            # single value
            if ":" not in val:
                match = op1.match(val)
                if not match:
                    sys.exit(errmsg + val + ")")
                op = match.group(2) if match.group(2) else defaultop
                operator = (
                    self.operators["negated"][op]
                    if match.group(1)
                    else self.operators["genuine"][op]
                )
                val = match.group(3)
                data[operator].add(val)

            # range
            else:
                match = op2.match(val)
                if not match:
                    sys.exit(errmsg + val + ")")
                operator = (
                    self.operators["negated"]["BETWEEN"]
                    if match.group(1)
                    else self.operators["genuine"]["BETWEEN"]
                )
                val = (match.group(2), match.group(3))
                data[operator].add(val)

        for operator, valset in data.items():
            condition, vals = self.get_conditional_expr(field, operator, *valset)
            conditions.extend(condition)
            vallist.extend(vals)

        return link.join(conditions), vallist

    def query_dates(self, field, *vals, link="OR"):
        link = " " + link.strip() + " "
        defaultop = "="
        op1 = re.compile(r"^(\^*)((?:>|>=|<|<=|!=|=)?)([0-9]{4}-[0-9]{2}-[0-9]{2})$")
        op2 = re.compile(
            r"^(\^*)([0-9]{4}-[0-9]{2}-[0-9]{2}):([0-9]{4}-[0-9]{2}-[0-9]{2})$"
        )
        errmsg = (
            "query error: date or date range expected for field " + field + " (got: "
        )
        data = defaultdict(set)
        conditions = []
        vallist = []

        for val in vals:
            val = str(val).strip()
            # single date
            if ":" not in val:
                match = op1.match(val)
                if not match:
                    sys.exit(errmsg + val + ")")
                op = match.group(2) if match.group(2) else defaultop
                operator = (
                    self.operators["negated"][op]
                    if match.group(1)
                    else self.operators["genuine"][op]
                )
                val = match.group(3)
                data[operator].add(val)

            # date range
            else:
                match = op2.match(val)
                if not match:
                    sys.exit(errmsg + val + ")")
                operator = (
                    self.operators["negated"]["BETWEEN"]
                    if match.group(1)
                    else self.operators["genuine"]["BETWEEN"]
                )
                val = (match.group(2), match.group(3))
                data[operator].add(val)

        for operator, valset in data.items():
            condition, vals = self.get_conditional_expr(field, operator, *valset)
            conditions.extend(condition)
            vallist.extend(vals)

        return link.join(conditions), vallist

    def query_string(self, field, *vals, link="OR"):
        link = " " + link.strip() + " "
        data = defaultdict(set)
        conditions = []
        vallist = []

        for val in vals:
            if val.startswith("^"):
                val = val[1:]
                opkey = "negated"
            else:
                opkey = "genuine"
            if val.startswith("%") or val.endswith("%"):
                operator = self.operators[opkey]["LIKE"]
            else:
                operator = self.operators[opkey]["="]

            data[operator].add(val)

        for operator, valset in data.items():
            condition, vals = self.get_conditional_expr(field, operator, *valset)
            conditions.extend(condition)
            vallist.extend(vals)

        return link.join(conditions), vallist

    def query_zip(self, field, *vals, link="OR"):
        link = " " + link.strip() + " "
        data = defaultdict(set)
        conditions = []
        vallist = []

        for val in vals:
            if val.startswith("^"):
                val = val[1:]
                opkey = "negated"
            else:
                opkey = "genuine"
            val = val.strip("%") + "%"
            operator = self.operators[opkey]["LIKE"]

            data[operator].add(val)

        for operator, valset in data.items():
            condition, vals = self.get_conditional_expr(field, operator, *valset)
            conditions.extend(condition)
            vallist.extend(vals)

        return link.join(conditions), vallist

    def query_metadata(self, name, *vals):
        conditions = ['"property_id" = ?']
        valueList = [self.properties[name]["id"]]
        targetfield = "value_" + self.properties[name]["datatype"]

        # query dates
        if self.properties[name]["querytype"] == "date":
            a, b = self.query_dates(targetfield, *vals)
            conditions.append(a)
            valueList.extend(b)

        # query numeric
        elif self.properties[name]["querytype"] == "numeric":
            a, b = self.query_numeric(targetfield, *vals)
            conditions.append(a)
            valueList.extend(b)

        # query text
        elif self.properties[name]["querytype"] == "text":
            a, b = self.query_string(targetfield, *vals)
            conditions.append(a)
            valueList.extend(b)

        # query zip
        elif self.properties[name]["querytype"] == "zip":
            a, b = self.query_zip(targetfield, *vals)
            conditions.append(a)
            valueList.extend(b)

        else:
            sys.exit(
                "error: unknown query type '"
                + self.properties[name]["querytype"]
                + "' for property '"
                + name
                + "'."
            )

        return (
            'SELECT "sample_id" AS id FROM sample2property WHERE '
            + " AND ".join(conditions),
            valueList,
        )

    def query_profile(self, *vars, reference_accession=None):  # noqa: C901
        iupac_nt_code = {
            "A": set("A"),
            "C": set("C"),
            "G": set("G"),
            "T": set("T"),
            "R": set("AGR"),
            "Y": set("CTY"),
            "S": set("GCS"),
            "W": set("ATW"),
            "K": set("GTK"),
            "M": set("ACM"),
            "B": set("CGTB"),
            "D": set("AGTD"),
            "H": set("ACTH"),
            "V": set("ACGV"),
            "N": set("ACGTRYSWKMBDHVN"),
        }
        iupac_aa_code = {
            "A": set("A"),
            "R": set("R"),
            "N": set("N"),
            "D": set("D"),
            "C": set("C"),
            "Q": set("Q"),
            "E": set("E"),
            "G": set("G"),
            "H": set("H"),
            "I": set("I"),
            "L": set("L"),
            "K": set("K"),
            "M": set("M"),
            "F": set("F"),
            "P": set("P"),
            "S": set("S"),
            "T": set("T"),
            "W": set("W"),
            "Y": set("Y"),
            "V": set("V"),
            "U": set("U"),
            "O": set("O"),
            "B": set("DNB"),
            "Z": set("EQZ"),
            "J": set("ILJ"),
            "Φ": set("VILFWYMΦ"),
            "Ω": set("FWYHΩ"),
            "Ψ": set("VILMΨ"),
            "π": set("PGASπ"),
            "ζ": set("STHNQEDKRζ"),
            "+": set("KRH+"),
            "-": set("DE-"),
            "X": set("ARNDCQEGHILKMFPSTWYVUOBZJΦΩΨπζ+-X"),
        }
        del_regex = re.compile(r"^(|[^:]+:)?([^:]+:)?del:(=?[0-9]+)(|-=?[0-9]+)?$")
        snv_regex = re.compile(r"^(|[^:]+:)?([^:]+:)?([A-Z]+)([0-9]+)(=?[A-Z]+)$")

        # set variants and generate sql
        base_sql = 'SELECT "sample.id" AS id FROM variantView WHERE '
        intersect_sqls = []
        intersect_vals = []
        except_sqls = []
        except_vals = []
        for var in vars:
            c = []  # condition
            v = []  # variable

            if var.startswith("^"):
                var = var[1:]
                negate = True
            else:
                negate = False

            # variant typing
            if match := snv_regex.match(var):
                snv = True
            elif match := del_regex.match(var):
                snv = False
            else:
                sys.exit("input error: " + var + " is not a valid variant definition.")

            # set molecule
            if match.group(1):
                c.append('"molecule.symbol" = ?')
                v.append(match.group(1)[:-1])
            else:
                c.append('"molecule.standard" = ?')
                v.append(1)

            # set element
            if match.group(2):
                c.append('"element.type" = ?')
                v.append("cds")
                c.append('"element.symbol" = ?')
                v.append(match.group(2)[:-1])
                code = iupac_aa_code
            else:
                c.append('"element.standard" = ?')
                v.append(1)
                code = iupac_nt_code

            # snp and insert handling
            if snv:
                c.append('"variant.start" = ?')
                v.append(int(match.group(4)) - 1)
                c.append('"variant.end" = ?')
                v.append(int(match.group(4)))
                c.append('"variant.ref" = ?')
                v.append(match.group(3))

                # explicit alternate allele
                if match.group(5).startswith("="):
                    c.append('"variant.alt" = ?')
                    v.append(match.group(5)[1:])

                # potentially ambiguous alternate snp
                elif len(match.group(5)) == 1:
                    l = len(code[match.group(5)])
                    if l == 1:
                        c.append('"variant.alt" = ?')
                        v.append(match.group(5))
                    else:
                        c.append('"variant.alt" IN (' + ", ".join(["?"] * l) + ")")
                        v.extend(code[match.group(5)])

                # potentially ambiguous alternate insert
                else:
                    a = [
                        "".join(x)
                        for x in itertools.product(*[code[x] for x in match.group(5)])
                    ]
                    l = len(a)
                    if l == 1:
                        c.append('"variant.alt" = ?')
                        v.extend(a)
                    else:
                        c.append('"variant.alt" IN (' + ", ".join(["?"] * l) + ")")
                        v.extend(a)

            # deletion handling
            else:
                s = match.group(3)
                e = match.group(4)[1:]

                if s.startswith("="):
                    s = s[1:]
                    c.append('"variant.start" = ?')
                else:
                    c.append('"variant.start" <= ?')

                if e.startswith("="):
                    e = e[1:]
                    c.append('"variant.end" = ?')
                else:
                    c.append('"variant.end" >= ?')
                v.append(int(s) - 1)
                v.append(int(e))

                c.append('"variant.alt" = ?')
                v.append(" ")

            # assemble sub-sql
            if negate:
                except_sqls.append(base_sql + " AND ".join(c))
                except_vals.extend(v)
            else:
                intersect_sqls.append(base_sql + " AND ".join(c))
                intersect_vals.extend(v)

        # assemble final sql
        if not intersect_sqls:
            intersect_sqls = [base_sql + "1"]

        sql = " INTERSECT ".join(intersect_sqls)

        if except_sqls:
            sql += " EXCEPT " + " EXCEPT ".join(except_sqls)

        return sql, intersect_vals + except_vals

    def count_molecules(self, reference_accession=None):
        if reference_accession:
            condition = '"reference.accession" = ?'
            vals = [reference_accession]
        else:
            condition = '"reference.standard" = ?'
            vals = [1]
        sql = (
            'SELECT count(DISTINCT "molecule.id") AS count FROM referenceView WHERE '
            + condition
        )
        return self.cursor.execute(sql, vals).fetchone()["count"]

    def match(  # noqa: 901
        self,
        *profiles,
        reserved_props=None,
        properties=None,
        reference_accession=None,
        showN=False,
        format="csv",
    ):
        # collecting sqls for metadata-based filtering
        property_sqls = []
        property_vals = []
        # IF sublineage search is enable
        # support: include and exclude
        if "with_sublineage" in reserved_props:
            _tmp_include_lin = []  # used to keep all lienages after search.
            lineage_col = reserved_props.get("with_sublineage")
            include_lin = properties.get(lineage_col)  # get list of given lineages
            negate = False
            print(include_lin)
            while include_lin:
                in_lin = include_lin.pop(0)

                if in_lin.startswith("^"):
                    in_lin = in_lin[1:]
                    negate = True

                # have wildcard in string which mean we have to find all lineage from wildcard query
                # then we used the wildcard query result to find all sublineages agian.
                if "%" in in_lin:
                    _tobeadded_lin = self.get_list_of_lineages(in_lin)
                    for i in _tobeadded_lin:

                        # if i != in_lin: # we dont need to add same lineage agian,so we skip for the duplicate lineage.
                        if negate:  # all lineage should add not ^
                            i = "^" + i
                        include_lin.append(i)  # add more lineage to find in next round.

                value = self.lineage_sublineage_dict.get(
                    in_lin, "none"
                )  # provide a default value if the key is missing:
                # print(value)
                if value != "none":
                    if negate:
                        in_lin = "^" + in_lin
                    _tmp_include_lin.append(in_lin)

                    _list = value.split(",")
                    for i in _list:
                        if negate:  # all sublineage should add not^
                            i = "^" + i
                        include_lin.append(i)  # add more lineage to find in next round.
                        # _tmp_include_lin.append(i)
                        # if we don't find this wildcard so we discard it
                else:  # None (no child)
                    if negate:
                        in_lin = "^" + in_lin
                    _tmp_include_lin.append(in_lin)
                negate = False

            include_lin = _tmp_include_lin
            properties[lineage_col] = include_lin
        # print(properties)

        if properties:
            for pname, vals in properties.items():
                sql, val = self.query_metadata(pname, *vals)
                property_sqls.append(sql)
                property_vals.extend(val)

        property_sqls = " INTERSECT ".join(property_sqls)

        # collecting sqls for genomic profile based filtering
        profile_sqls = []
        profile_vals = []
        for profile in profiles:
            sql, val = self.query_profile(
                *profile, reference_accession=reference_accession
            )
            profile_sqls.append(sql)
            profile_vals.extend(val)

        if len(profiles) == 1:
            profile_sqls = profile_sqls[0]
        elif len(profiles) > 1:
            profile_sqls = " UNION ".join(
                ["SELECT * FROM (" + x + ")" for x in profile_sqls]
            )
        else:
            profile_sqls = ""

        if property_sqls and profile_sqls:
            if len(profiles) > 1:
                sample_selection_sql = (
                    property_sqls + " INTERSECT SELECT * FROM (" + profile_sqls + ")"
                )
            else:
                sample_selection_sql = property_sqls + " INTERSECT " + profile_sqls
        elif property_sqls or profile_sqls:
            sample_selection_sql = property_sqls + profile_sqls
        else:
            if "sample" in reserved_props:
                # if 'sample' is presented we just use only samples
                samples_condition = []
                for pname, vals in reserved_props.items():
                    if pname == "sample":
                        for x in vals:
                            samples_condition.append('"' + x + '"')
                sample_selection_sql = (
                    "SELECT id FROM sample WHERE name IN ("
                    + " , ".join(samples_condition)
                    + ")"
                )

                property_sqls = []
                property_vals = []
            else:
                sample_selection_sql = "SELECT id FROM sample"

        genome_element_condition = [
            str(x) for x in self.get_element_ids(reference_accession, "source")
        ]
        if len(genome_element_condition) == 1:
            genome_element_condition = '"element.id" = ' + genome_element_condition[0]
            m = ""
        else:
            genome_element_condition = (
                '"element.id" IN (' + ", ".join(genome_element_condition) + ")"
            )
            m = ' "molecule.symbol" || "@" || '

        if not showN:
            nn = ' AND "variant.alt" != "N" '
            nx = ' AND "variant.alt" != "X" '
        else:
            nn = ""
            nx = ""

        cds_element_condition = [
            str(x) for x in self.get_element_ids(reference_accession, "cds")
        ]
        if len(cds_element_condition) == 1:
            cds_element_condition = '"element.id" = ' + cds_element_condition[0]
        else:
            cds_element_condition = (
                '"element.id" IN (' + ", ".join(cds_element_condition) + ")"
            )

        # standard query
        if format == "csv" or format == "tsv":

            # select samples
            sql = sample_selection_sql
            sample_ids = self.cursor.execute(
                sql, property_vals + profile_vals
            ).fetchall()

            if not sample_ids:
                return []
            selected_sample_ids = ", ".join([str(x["id"]) for x in sample_ids])
            # rows = {x['id']: {"id": x['id']} for x in sample_ids}
            # print(len(sample_ids))

            #
            # Current solution:
            # After we got the selected IDs
            # We use two-stage query and then combine both results together to produce final result
            # 1. Query properties
            # 2. Query  AA/NT profile
            fields = ['"sample.name"'] + ['"' + x + '"' for x in self.properties]
            sql = "SELECT name as " + ", ".join(fields) + "FROM sample "

            joins = [
                "LEFT JOIN (SELECT sample_id, value_"
                + y["datatype"]
                + " as "
                + x
                + " FROM sample2property WHERE property_id = "
                + str(y["id"])
                + ") as t"
                + str(y["id"])
                + " ON sample.id = t"
                + str(y["id"])
                + ".sample_id"
                for x, y in self.properties.items()
            ]
            _1_final_sql = (
                sql
                + " ".join(joins)
                + " WHERE sample.id IN ("
                + selected_sample_ids
                + ")"
            )
            # print(_1_final_sql)
            _1_rows = self.cursor.execute(_1_final_sql).fetchall()
            _2_final_sql = (
                " SELECT name AS 'sample.name'  , nt_profile._profile AS NUC_PROFILE, aa_profile._profile AS AA_PROFILE \
                    FROM \
                            ( \
                              SELECT  \"sample.id\", group_concat("
                + m
                + '"variant.label") AS _profile \
                              FROM variantView WHERE "sample.id" IN ('
                + selected_sample_ids
                + ") AND "
                + genome_element_condition
                + nn
                + ' GROUP BY "sample.id" \
                            ) nt_profile, \
                            ( \
                              SELECT  "sample.id", group_concat('
                + m
                + '"element.symbol" || ":" || "variant.label") AS _profile \
                              FROM variantView WHERE "sample.id" IN ('
                + selected_sample_ids
                + ") AND "
                + cds_element_condition
                + nx
                + ' GROUP BY "sample.id" \
                            ) aa_profile, \
                            sample \
                    WHERE nt_profile."sample.id" = aa_profile."sample.id" AND  nt_profile."sample.id" = sample.id  \
                          AND sample.id  IN ('
                + selected_sample_ids
                + ")"
            )
            _2_rows = self.cursor.execute(_2_final_sql).fetchall()
            if len(_1_rows) != len(_2_rows):
                print("WARNING: There are something error between query")
            # print(set([ x['sample.name'] for x in _1_rows ]) ^ set([ x['name'] for x in _2_rows ]))
            # To combine:
            # We update list of dict (update on result from query #2)
            # merge all results
            _1_rows.extend(
                list(
                    map(
                        lambda x, y: x.update(
                            {
                                key: value
                                for key, value in y.items()
                                if (key == "NUC_PROFILE") or (key == "AA_PROFILE")
                            }
                        )
                        if x.get("sample.name") == y.get("sample.name")
                        else None,
                        _1_rows,
                        _2_rows,
                    )
                )
            )

            _1_rows = list(filter(None, _1_rows))

            # since we use "update" function (i.e. extends the dict. to include all key:value from properties base on sample name)
            # at _1_rows so we can return _1_rows only a
            return _1_rows  # list(rows.values())
            """
            sql = "WITH selected_samples AS (" + sample_selection_sql + ") \
               SELECT  *, \
                        ( \
                          SELECT group_concat(" + m + "\"variant.label\") AS nuc_profile \
                          FROM variantView WHERE \"sample.id\" IN (SELECT id FROM selected_samples) AND " + genome_element_condition + nn + " GROUP BY \"sample.name\" ORDER BY \"element.id\", \"variant.start\" \
                        ) nt_profile, \
                        ( \
                          SELECT group_concat(" + m + "\"element.symbol\" || \":\" || \"variant.label\") AS nuc_profile \
                          FROM variantView WHERE \"sample.id\" IN (SELECT id FROM selected_samples) AND " + cds_element_condition + np + " GROUP BY \"sample.name\" ORDER BY \"element.id\", \"variant.start\" \
                        ) aa_profile \
                FROM sample \
                WHERE id IN ( SELECT id FROM selected_samples )"


            sql = "SELECT * \
                   FROM variantView \
                   WHERE \"sample.id\" IN (" + sample_selection_sql + ")" # ORDER BY \"sample.id\", \"element.id\", \"variant.start\""
            """
        elif format == "count":
            sql = (
                "SELECT count(DISTINCT id) as count FROM (" + sample_selection_sql + ")"
            )
        elif format == "vcf":
            sql = (
                'SELECT "element.id",  "element.type", "molecule.accession", "variant.start", "variant.ref", "variant.alt", "variant.label", group_concat("sample.name") as samples FROM variantView WHERE "sample.id" IN ('
                + sample_selection_sql
                + ") AND "
                + genome_element_condition
                + nn
                + 'GROUP BY "molecule.accession", "variant.start", "variant.ref", "variant.alt" ORDER BY "molecule.accession", "variant.start"'
            )
        else:
            sys.exit("error: '" + format + "' is not a valid output format")
        return self.cursor.execute(sql, property_vals + profile_vals)  # .fetchall()

    def get_list_of_lineages(self, lineage):
        sql = (
            "SELECT DISTINCT lineage FROM lineages WHERE lineage LIKE '"
            + lineage
            + "';"
        )
        rows = self.cursor.execute(sql).fetchall()
        result = [i["lineage"] for i in rows]
        return result

    # MISC

    @staticmethod
    def optimize(dbfile):
        with sqlite3.connect(dbfile) as con:
            con.executescript("PRAGMA analysis_limit=400;")
            con.executescript("PRAGMA optimize;")
            con.executescript("VACUUM;")

    @staticmethod
    def dict_factory(cursor, row):
        d = {}  # OrderedDict()
        for idx, col in enumerate(cursor.description):
            d[col[0]] = row[idx]
        return d

    # Utils.
    def get_db_size(self, decimal_places=3):
        size = os.path.getsize(self.dbfile)
        for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
            if size < 1024.0:
                break
            size /= 1024.0
        return f"{size:.{decimal_places}f}{unit}"

    @staticmethod
    def upgrade_db(dbfile):
        try:
            with sqlite3.connect(dbfile) as con:
                cur = con.cursor()
                current_version = cur.execute("pragma user_version").fetchone()[0]

            print(
                "Current version:",
                current_version,
                " Upgrade to:",
                SUPPORTED_DB_VERSION,
            )
            uri = "file:" + urlquote(dbfile)
            print("Perform the Upgrade:", uri)
            while current_version < SUPPORTED_DB_VERSION:

                next_version = current_version + 1
                file_path = os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "migrate/" + str(next_version) + ".sql",
                )
                if not os.path.isfile(file_path):
                    raise ValueError("Sorry, we cannot find", file_path)

                with open(file_path, "r") as handle:
                    sql = handle.read()
                with sqlite3.connect(uri + "?mode=rwc", uri=True) as con:
                    con.executescript(sql)

                current_version = next_version

        except sqlite3.Error as er:
            con.executescript("ROLLBACK")
            raise er
        except ValueError as er:
            print(er)
        finally:
            print("Database now version:", current_version)
            if current_version == SUPPORTED_DB_VERSION:
                print("Success: Database upgrade was successfully completed")
            else:
                print("Error: Upgrade was not completed")



#!/usr/bin/python
# Maintainer: KongkitimanonK
# Script version 1 (copy from VOCAL)
# The method originally came from  https://github.com/cov-lineages/pango-designation.
# We just adapt and change some parts to be used in covsonar, vocal etc.

import json
import os
import shutil
from tempfile import mkdtemp

import pandas as pd
import requests


class Aliasor:
    def __init__(self, alias_file):
        aliases = pd.read_json(alias_file)
        self.alias_dict = {
            x: x if x.startswith("X") else aliases[x][0] for x in aliases.columns
        }
        self.alias_dict["A"] = "A"
        self.alias_dict["B"] = "B"
        self.realias_dict = {v: k for k, v in self.alias_dict.items()}

    def compress(self, name):
        name_split = name.split(".")
        if len(name_split) < 5:
            return name
        letter = self.realias_dict[".".join(name_split[0:4])]
        if len(name_split) == 5:
            return letter + "." + name_split[4]
        else:
            return letter + "." + ".".join(name_split[4:])

    def uncompress(self, name):
        name_split = name.split(".")
        letter = name_split[0]
        unaliased = self.alias_dict[letter]
        if len(name_split) == 1:
            return name
        if len(name_split) == 2:
            return unaliased + "." + name_split[1]
        else:
            return unaliased + "." + ".".join(name_split[1:])


class sonarLinmgr:
    def __init__(self, tmpdir=None):
        self._tmpdir = mkdtemp(prefix=".tmp_sonarLinmgr_")
        self._linurl = "https://raw.githubusercontent.com/cov-lineages/pango-designation/master/lineages.csv"
        self._aliurl = "https://raw.githubusercontent.com/cov-lineages/pango-designation/master/pango_designation/alias_key.json"
        self.lineage_file = os.path.join(self._tmpdir, "lineages.csv")
        self.alias_file = os.path.join(self._tmpdir, "alias.json")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        # print('Clean env.')
        shutil.rmtree(self._tmpdir)

    # download
    def download_lineage_data(self):
        url_content = requests.get(self._linurl)
        with open(self.lineage_file, "wb") as handle:
            handle.write(url_content.content)
        items = requests.get(self._aliurl)
        with open(self.alias_file, "w") as handle:
            json.dump(items.json(), handle)

    # processing
    @staticmethod
    def lts(lineage):
        items = []
        for item in lineage.split("."):
            item_string = str(item)
            items.append((5 - len(item)) * "0" + item_string)
        return "".join(items)

    def process_lineage_data(self, output_file):
        # handle duplicate values
        with open(self.alias_file) as f:
            # load json objects to dictionaries
            data_dict = json.load(f)
        for k, v in data_dict.items():
            if type(v) is list:
                data_dict[k] = list(set(v))
        # rewrite the json
        with open(self.alias_file, "w") as nf:
            json.dump(data_dict, nf)

        aliasor = Aliasor(self.alias_file)
        df_lineages = pd.read_csv(self.lineage_file)
        lineages = df_lineages.lineage.unique()

        uncompressed_lineages = []
        sorted_lineages = []

        # Calculating parent-child relationship
        uncompressed_lineages = [x for x in map(aliasor.uncompress, lineages)]
        uncompressed_lineages.sort(key=sonarLinmgr.lts)
        sorted_lineages = [x for x in map(aliasor.compress, uncompressed_lineages)]

        datadict = []
        for _id in lineages:
            sub_lineage_char = aliasor.realias_dict.get(_id)
            sub_lineage_list = [
                x for x in sorted_lineages if x.split(".")[0] == sub_lineage_char
            ]
            sub_lineage_list = list(filter((_id).__ne__, sub_lineage_list))
            if len(sub_lineage_list):
                datadict.append(
                    {"lineage": _id, "sublineage": ",".join(sub_lineage_list)}
                )
            else:
                datadict.append({"lineage": _id, "sublineage": "none"})

        return pd.DataFrame(
            datadict
        )  # pd.DataFrame(datadict).to_csv(output_file, sep="	", index=False)

    def update_lineage_data(self, output_file):
        self.download_lineage_data()
        df = self.process_lineage_data(output_file)
        return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# author: Stephan Fuchs (Robert Koch Institute, MF1, fuchss@rki.de)

# DEPENDENCIES
import collections
import csv
import gzip
import logging
import lzma
import os
import re
import sys
from tempfile import mkstemp

from Bio import SeqIO
from Bio.SeqUtils.CheckSum import seguid
import pandas as pd

from . import __version__
from .dbm import sonarDBManager


class Aliasor:
    def __init__(self, alias_file):
        aliases = pd.read_json(alias_file)
        self.alias_dict = {
            x: x if x.startswith("X") else aliases[x][0] for x in aliases.columns
        }
        self.alias_dict["A"] = "A"
        self.alias_dict["B"] = "B"
        self.realias_dict = {v: k for k, v in self.alias_dict.items()}

    def compress(self, name):
        name_split = name.split(".")
        if len(name_split) < 5:
            return name
        letter = self.realias_dict[".".join(name_split[0:4])]
        if len(name_split) == 5:
            return letter + "." + name_split[4]
        else:
            return letter + "." + ".".join(name_split[4:])

    def uncompress(self, name):
        name_split = name.split(".")
        letter = name_split[0]
        unaliased = self.alias_dict[letter]
        if len(name_split) == 1:
            return name
        if len(name_split) == 2:
            return unaliased + "." + name_split[1]
        else:
            return unaliased + "." + ".".join(name_split[1:])


# CLASS
class sonarBasics(object):
    """
    this object provides sonarBasics functionalities and intelligence

    Notes
    -----
            Please note, that genomic and protein coordinates are expected to be  and
            returned 0-based by this object, except for formatted profiles.
            While start or single coordinates are inclusive, end coordinates of
            ranges are exclusive, expressed in a mathematical notation: [start, end).
            Only in formatted profiles start and end coordinates are 1-based and both
            inclusive.

    Examples
    --------

    In this example the path to the database is stored in DOCTESTDB.

    >>> db = sonarBasics(DOCTESTDB)

    Parameters
    ----------
    dbfile : str
            define a path to a non-existent or valid SONAR database file. If the
            file does not exist, a SONAR database is created.
    translation_table : int
            define the genetic code table used for in silico translation (see
            https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi) [ 1 ]

    Attributes
    ----------
    db : str
            stores the absolute path to the used SONAR database file
    reffna : str
            stores the absolute path to the built-in FASTA file containing the reference
            genome sequence
    refgff : str
            stores the absolute path to the built-in GFF3 file containing the reference
            genome annotation
    translation_table : int
            stores the genetic code table used for in silico translation (see
            https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi) [ 1 ]
    refseq : str
            stores the upper-case sequence of the built-in reference genome
    refdescr : str
            stores the FASTA header of the built-in reference genome
    refgffObj : object
            stores the sonarGFF object based on the built-in reference genome
            annotation
    iupac_nt_code : dict
            stores a dict with IUPAC one-letter nucleotide codes as keys and the
            respective set of matching explicit IUPAC one-letter nucleotide codes
            as values (e.g {"W": set('A', 'T')})
    iupac_explicit_nt_code : dict
            stores a set containing all non-ambiguous IUPAC one-letter nucleotide codes
    iupac_ambig_nt_code : set
            stores a set containing all ambiguous IUPAC one-letter nucleotide codes
    iupac_aa_code : dict
            stores a dict with IUPAC one-letter amino acid codes as keys and
            the respective set of matching IUPAC one-letter amino acids codes as values
    iupac_explicit_aa_code : dict
            stores a set containing all non-ambiguous IUPAC one-letter amino acid codes
    iupac_ambig_aa_code : dict
            stores a set containing all ambiguous IUPAC one-letter amino acid codes
    dna_var_regex : compiled re expression
            stores a compiled re expression that matches to nucleotide profiles but
            not to amino acid profiles
    aa_var_regex : compiled re expression
            stores a compiled re expression that matches to amino acid profiles but
            not to nucleotide profiles
    del_regex : compiled re expression
            stores a compiled re expression that matches to deletion profiles on
            nucleotide as well as on amino acid level.
    dnavar_grep_regex : compiled re expression
            stores a compiled re expression that matches to snp or dna insertion
            profiles with eference allele, genomic position and variant allele
            as groups.
    codedict : dict
            stores a dictionary with "dna" and "aa" containing the field name in the
            database that stores the profile data, the one letter code with and
            without ambiguities
    """

    def __init__(self, dbfile, translation_table=1, debug=False):
        self.db = os.path.abspath(dbfile) if dbfile else mkstemp()[1]
        self.debug = debug
        self.__moduledir = self.get_module_base()
        self.reffna = os.path.join(self.__moduledir, "ref.fna")
        self.refgff = os.path.join(self.__moduledir, "ref.gff3")
        self.translation_table = translation_table
        self.__refseq = None
        self.__refdescr = None
        self.__refgffObj = None
        self.__iupac_nt_code = None
        self.__iupac_aa_code = None
        self.__iupac_explicit_nt_code = None
        self.__iupac_explicit_aa_code = None
        self.__iupac_ambig_nt_code = None
        self.__iupac_ambig_aa_code = None
        self.__terminal_letters_regex = re.compile("[A-Z]$")
        self.__dna_var_regex = None
        self.__aa_var_regex = None
        self.__del_regex = None
        self.__dnavar_grep_regex = None
        self.__codedict = None
        self.__fasta_tag_regex = None
        self._covSonar_version = self.get_version()
        logging.basicConfig(format="%(asctime)s %(message)s")

    @staticmethod
    def get_version():
        return __version__

    # DB MAINTENANCE

    @staticmethod
    def get_module_base(*join_with):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), *join_with)

    @staticmethod
    def setup_db(  # noqa: C901
        fname, auto_create=False, default_setup=True, reference_gb=None, debug=False
    ):
        if os.path.isfile(fname):
            sys.exit("setup error: " + fname + " does already exist.")
        sonarDBManager.setup(fname, debug=debug)

        # loading default data
        if default_setup or auto_create:
            with sonarDBManager(fname, readonly=False, debug=debug) as dbm:
                # adding pre-defined sample properties
                dbm.add_property(
                    "imported",
                    "date",
                    "date",
                    "date sample has been imported to the database",
                )
                dbm.add_property(
                    "modified",
                    "date",
                    "date",
                    "date when sample data has been modified lastly",
                )
                dbm.add_property(
                    "nuc_profile",
                    "text",
                    "text",
                    "stores the nucleotide level profiles",
                )
                dbm.add_property(
                    "aa_profile", "text", "text", "stores the aa level profiles"
                )
                dbm.add_property(
                    "nuc_n_profile",
                    "text",
                    "text",
                    "stores the nucleotide(with N) level profiles",
                )
                dbm.add_property(
                    "aa_x_profile",
                    "text",
                    "text",
                    "stores the aa(with X) level profiles",
                )

                # if enable, create PREDEFINED properties
                if auto_create:
                    dbm.add_property("DATE_DRAW", "date", "date", "Sampling date")
                    dbm.add_property(
                        "PROCESSING_DATE", "date", "date", "Submission/Processing date"
                    )
                    dbm.add_property(
                        "COUNTRY", "text", "text", "Country where a sample belongs to"
                    )
                    dbm.add_property("HOST", "text", "text", "e.g., HUMAN")
                    dbm.add_property("ZIP", "text", "text", "zip code e.g., 33602")
                    dbm.add_property("LAB", "text", "text", "lab id e.g., 11069")
                    dbm.add_property("LINEAGE", "text", "text", "e.g., BA.2 or B.1.1.7")
                    dbm.add_property("TECHNOLOGY", "text", "text", "e.g., ILLUMINA")

                # adding reference
                if not reference_gb:
                    reference_gb = os.path.join(
                        os.path.dirname(os.path.abspath(__file__)), "data/ref.gb"
                    )
                records = [x for x in sonarBasics.iter_genbank(reference_gb)]
                ref_id = dbm.add_reference(
                    records[0]["accession"],
                    records[0]["description"],
                    records[0]["organism"],
                    1,
                    1,
                )

                # adding reference molecule and elements
                for i, record in enumerate(records):
                    gene_ids = {}
                    s = 1 if i == 0 else 0

                    mol_id = dbm.insert_molecule(
                        ref_id,
                        record["moltype"],
                        record["accession"],
                        record["symbol"],
                        record["description"],
                        i,
                        record["length"],
                        s,
                    )

                    # source handling
                    source_id = dbm.insert_element(
                        mol_id,
                        "source",
                        record["source"]["accession"],
                        record["source"]["symbol"],
                        record["source"]["description"],
                        record["source"]["start"],
                        record["source"]["end"],
                        record["source"]["strand"],
                        record["source"]["sequence"],
                        standard=1,
                        parts=record["source"]["parts"],
                    )
                    if record["source"]["sequence"] != dbm.get_sequence(source_id):
                        sys.exit(
                            "genbank error: could not recover sequence of '"
                            + record["source"]["accession"]
                            + "' (source)"
                        )

                    # gene handling
                    for elem in record["gene"]:
                        gene_ids[elem["accession"]] = dbm.insert_element(
                            mol_id,
                            "gene",
                            elem["accession"],
                            elem["symbol"],
                            elem["description"],
                            elem["start"],
                            elem["end"],
                            elem["strand"],
                            elem["sequence"],
                            standard=0,
                            parent_id=source_id,
                            parts=elem["parts"],
                        )
                        if elem["sequence"] != dbm.extract_sequence(
                            gene_ids[elem["accession"]]
                        ):
                            sys.exit(
                                "genbank error: could not recover sequence of '"
                                + elem["accession"]
                                + "' (gene)"
                            )

                    # cds handling
                    for elem in record["cds"]:
                        cid = dbm.insert_element(
                            mol_id,
                            "cds",
                            elem["accession"],
                            elem["symbol"],
                            elem["description"],
                            elem["start"],
                            elem["end"],
                            elem["strand"],
                            elem["sequence"],
                            0,
                            gene_ids[elem["gene"]],
                            elem["parts"],
                        )
                        if elem["sequence"] != dbm.extract_sequence(
                            cid, translation_table=1
                        ):
                            sys.exit(
                                "genbank error: could not recover sequence of '"
                                + elem["accession"]
                                + "' (cds)"
                            )
                print("Success: Database was successfully installed")

    # DATA IMPORT
    # genbank handling handling
    @staticmethod
    def process_segments(feat_location_parts, cds=False):
        base = 0
        div = 1 if not cds else 3
        segments = []
        for i, segment in enumerate(feat_location_parts, 1):
            segments.append(
                [int(segment.start), int(segment.end), segment.strand, base, i]
            )
            base += round((segment.end - segment.start - 1) / div, 1)
        return segments

    @staticmethod
    def iter_genbank(fname):
        gb_data = {}
        for gb_record in SeqIO.parse(fname, "genbank"):
            # adding general annotation
            gb_data["accession"] = (
                gb_record.name + "." + str(gb_record.annotations["sequence_version"])
            )
            gb_data["symbol"] = (
                gb_record.annotations["symbol"]
                if "symbol" in gb_record.annotations
                else ""
            )
            gb_data["organism"] = gb_record.annotations["organism"]
            gb_data["moltype"] = None
            gb_data["description"] = gb_record.description
            gb_data["length"] = None
            gb_data["segment"] = ""
            gb_data["gene"] = []
            gb_data["cds"] = []
            gb_data["source"] = None

            # adding source annotation
            source = [x for x in gb_record.features if x.type == "source"]
            if len(source) != 1:
                sys.exit(
                    "genbank error: expecting exactly one source feature (got: "
                    + str(len(source))
                    + ")"
                )
            feat = source[0]
            gb_data["moltype"] = (
                feat.qualifiers["mol_type"][0] if "mol_type" in feat.qualifiers else ""
            )
            gb_data["source"] = {
                "accession": gb_data["accession"],
                "symbol": gb_data["accession"],
                "start": int(feat.location.start),
                "end": int(feat.location.end),
                "strand": "",
                "sequence": sonarBasics.harmonize(feat.extract(gb_record.seq)),
                "description": "",
                "parts": sonarBasics.process_segments(feat.location.parts),
            }
            gb_data["length"] = len(gb_data["source"]["sequence"])
            if "segment" in feat.qualifiers:
                gb_data["segment"] = feat.qualifiers["segment"][0]

            for feat in gb_record.features:
                # adding gene annotation
                if feat.type == "gene":
                    gb_data["gene"].append(
                        {
                            "accession": feat.id
                            if feat.id != "<unknown id>"
                            else feat.qualifiers["gene"][0],
                            "symbol": feat.qualifiers["gene"][0],
                            "start": int(feat.location.start),
                            "end": int(feat.location.end),
                            "strand": feat.strand,
                            "sequence": sonarBasics.harmonize(
                                feat.extract(gb_data["source"]["sequence"])
                            ),
                            "description": "",
                            "parts": sonarBasics.process_segments(feat.location.parts),
                        }
                    )

                # adding cds annotation
                elif feat.type == "CDS":
                    gb_data["cds"].append(
                        {
                            "accession": feat.qualifiers["protein_id"][0],
                            "symbol": feat.qualifiers["gene"][0],
                            "start": int(feat.location.start),
                            "end": int(feat.location.end),
                            "strand": feat.strand,
                            "gene": feat.qualifiers["gene"][0],
                            "sequence": feat.qualifiers["translation"][0],
                            "description": feat.qualifiers["product"][0],
                            "parts": sonarBasics.process_segments(
                                feat.location.parts, True
                            ),
                        }
                    )
            yield gb_data

    # seq handling
    @staticmethod
    def hash(seq):
        """ """
        return seguid(seq)

    @staticmethod
    def harmonize(seq):
        """ """
        return str(seq).strip().upper().replace("U", "T")

    # matching
    @staticmethod
    def match(
        db,
        profiles=[],
        reserved_props_dict={},
        propdict={},
        reference=None,
        outfile=None,
        format="csv",
        debug="False",
    ):
        with sonarDBManager(db, debug=debug) as dbm:
            if format == "vcf" and reference is None:
                reference = dbm.get_default_reference_accession()
            cursor = dbm.match(
                *profiles,
                reserved_props=reserved_props_dict,
                properties=propdict,
                reference_accession=reference,
                format=format,
            )
            if format == "csv" or format == "tsv":
                tsv = True if format == "tsv" else False
                sonarBasics.exportCSV(
                    cursor, outfile=outfile, na="*** no match ***", tsv=tsv
                )
            elif format == "count":
                if outfile:
                    with open(outfile, "w") as handle:
                        handle.write(str(cursor.fetchone()["count"]))
                else:
                    print(cursor.fetchone()["count"])
            elif format == "vcf":
                sonarBasics.exportVCF(
                    cursor, reference=reference, outfile=outfile, na="*** no match ***"
                )
            else:
                sys.exit("error: '" + format + "' is not a valid output format")

    # restore
    @staticmethod
    def restore(
        db, *samples, reference_accession=None, aligned=False, outfile=None, debug=False
    ):
        with sonarDBManager(db, readonly=True, debug=debug) as dbm:
            handle = sys.stdout if outfile is None else open(outfile, "w")
            for sample in samples:
                prefixes = collections.defaultdict(str)
                molecules = {
                    x["element.id"]: {
                        "seq": list(x["element.sequence"]),
                        "mol": x["element.symbol"],
                    }
                    for x in dbm.get_alignment_data(sample, reference_accession=None)
                }
                gap = "-" if aligned else ""
                for vardata in dbm.iter_dna_variants(sample, *molecules.keys()):
                    if aligned and len(vardata["variant.alt"]) > 1:
                        vardata["variant.alt"] = (
                            vardata["variant.alt"][0]
                            + vardata["variant.alt"][1:].lower()
                        )
                    if vardata["variant.alt"] == " ":
                        for i in range(
                            vardata["variant.start"], vardata["variant.end"]
                        ):
                            molecules[vardata["element.id"]]["seq"][i] = gap
                    elif vardata["variant.start"] >= 0:
                        molecules[vardata["element.id"]]["seq"][
                            vardata["variant.start"]
                        ] = vardata["variant.alt"]
                    else:
                        prefixes[vardata["element.id"]] = vardata["variant.alt"]
                molecules_len = len(molecules)
                records = []
                for element_id in molecules:
                    if molecules_len == 1:
                        records.append(">" + sample)
                    else:
                        records.append(
                            ">"
                            + sample
                            + " [molecule="
                            + molecules[vardata["element.id"]]["mol"]
                            + "]"
                        )
                    records.append(
                        prefixes[element_id]
                        + "".join(molecules[vardata["element.id"]]["seq"])
                    )
                if len(records) > 0:
                    handle.write("\n".join(records) + "\n")

    # restore
    @staticmethod
    def delete(db, *samples, debug):
        with sonarDBManager(db, readonly=False, debug=debug) as dbm:
            before = dbm.count_samples()
            dbm.delete_samples(*samples)
            after = dbm.count_samples()
            print(
                before - after,
                "of",
                len(samples),
                "samples found and deleted.",
                after,
                "samples remain in the database.",
            )

    @staticmethod
    def show_db_info(db):
        with sonarDBManager(db, readonly=True) as dbm:
            print("covSonar Version:          ", sonarBasics.get_version())
            print("database path:             ", dbm.dbfile)
            print("database version:          ", dbm.get_db_version())
            print("database size:             ", dbm.get_db_size())
            print("unique sequences:          ", dbm.count_sequences())
            # print("Sample properties          ", dbm.get_earliest_import())
            # print("latest genome import:      ", dbm.get_latest_import())
            # print("earliest sampling date:    ", dbm.get_earliest_date())
            # print("latest sampling date:      ", dbm.get_latest_date())

    # output
    # profile generation
    @staticmethod
    def iter_formatted_match(cursor):
        nuc_profiles = collections.defaultdict(list)
        aa_profiles = collections.defaultdict(list)
        samples = set()
        logging.warning("getting profile data")
        logging.warning("processing profile data")
        for row in cursor:

            samples.add(row["samples"])
            if row["element.type"] == "cds":
                aa_profiles[row["samples"]].append(
                    (
                        row["element.id"],
                        row["variant.start"],
                        row["element.symbol"] + ":" + row["variant.label"],
                    )
                )
            else:
                nuc_profiles[row["samples"]].append(
                    (row["element.id"], row["variant.start"], row["variant.label"])
                )
        out = []
        print(len(samples))
        for sample in sorted(samples):
            print(sorted(nuc_profiles[sample], key=lambda x: (x[0], x[1])))

        logging.warning("assembling profile data")
        for sample in sorted(samples):
            print(sample)
            out.append(
                {
                    "sample.name": sample,
                    "nuc_profile": " ".join(
                        sorted(nuc_profiles[sample], key=lambda x: (x[0], x[1]))
                    ),
                    "aa_profile": " ".join(
                        sorted(aa_profiles[sample], key=lambda x: (x[0], x[1]))
                    ),
                }
            )
        print(out)
        return out

    # csv
    def exportCSV(cursor, outfile=None, na="*** no data ***", tsv=False):
        i = -1
        try:
            for i, row in enumerate(cursor):
                if i == 0:
                    outfile = sys.stdout if outfile is None else open(outfile, "w")
                    sep = "\t" if tsv else ","
                    writer = csv.DictWriter(
                        outfile, row.keys(), delimiter=sep, lineterminator=os.linesep
                    )  # extrasaction='ignore',
                    writer.writeheader()
                writer.writerow(row)
            if i == -1:
                print(na)
        except Exception:
            print("An exception occurred", row)
            raise

    # vcf
    def exportVCF(cursor, reference, outfile=None, na="*** no match ***"):  # noqa: C901
        records = collections.OrderedDict()
        all_samples = set()
        for row in cursor.fetchall():  # sonarBasics.iter_formatted_match(cursor):
            chrom, pos, ref, alt, samples = (
                row["molecule.accession"],
                row["variant.start"],
                row["variant.ref"],
                row["variant.alt"],
                row["samples"],
            )
            print(samples)
            if chrom not in records:
                records[chrom] = collections.OrderedDict()
            if pos not in records[chrom]:
                records[chrom][pos] = {}
            if ref not in records[chrom][pos]:
                records[chrom][pos][ref] = {}
            records[chrom][pos][ref][alt] = set(samples.split("\t"))
            all_samples.update(samples.split("\t"))
        print(all_samples)
        if len(records) != 0:
            all_samples = sorted(all_samples)
            if outfile is None:
                handle = sys.stdout
            else:
                # if not outfile.endswith(".gz"):
                # outfile += ".gz"
                handle = open(outfile, mode="w")  # bgzf.open(outfile, mode='wb')

            # vcf header
            handle.write("##fileformat=VCFv4.2\n")
            handle.write("##poweredby=covSonar2\n")
            handle.write("##reference=" + reference + "\n")
            handle.write(
                '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype"\n'
            )
            handle.write(
                "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t"
                + "\t".join(all_samples)
                + "\n"
            )

            # records
            for chrom in records:
                for pos in records[chrom]:
                    for ref in records[chrom][pos]:
                        # snps and inserts (combined output)
                        alts = [x for x in records[chrom][pos][ref].keys() if x.strip()]
                        if alts:
                            alt_samples = set()
                            gts = []
                            for alt in alts:
                                samples = records[chrom][pos][ref][alt]
                                alt_samples.update(samples)
                                gts.append(
                                    ["1" if x in samples else "0" for x in all_samples]
                                )
                            gts = [
                                ["0" if x in alt_samples else "1" for x in all_samples]
                            ] + gts
                            record = [
                                chrom,
                                str(pos),
                                ref,
                                ",".join(alts),
                                ".",
                                ".",
                                ".",
                                "GT",
                            ] + ["/".join(x) for x in zip(*gts)]
                        # dels (individual output)
                        for alt in [
                            x for x in records[chrom][pos][ref].keys() if not x.strip()
                        ]:
                            samples = records[chrom][pos][ref][alt]
                            record = [
                                chrom,
                                str(pos),
                                ref,
                                ref[0],
                                ".",
                                ".",
                                ".",
                                "GT",
                            ] + ["0/1" if x in samples else "1/0" for x in all_samples]
                        handle.write("\t".join(record) + "\n")
            handle.close()
        else:
            print(na)

    # gerenal
    @staticmethod
    def open_file(fname, mode="r", compressed=False, encoding=None):
        if not os.path.isfile(fname):
            sys.exit("input error: " + fname + " does not exist.")
        if compressed == "auto":
            compressed = os.path.splitext(fname)[1][1:]
        try:
            if compressed == "gz":
                return gzip.open(fname, mode + "t", encoding=encoding)
            if compressed == "xz":
                return lzma.open(fname, mode + "t", encoding=encoding)
            else:
                return open(fname, mode, encoding=encoding)
        except Exception:
            sys.exit("input error: " + fname + " cannot be opened.")

    @staticmethod
    def set_key(dictionary, key, value):
        if key not in dictionary:
            dictionary[key] = value
        elif type(dictionary[key]) == list:
            dictionary[key].append(value)
        else:
            dictionary[key] = [dictionary[key], value]
        return dictionary



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# author: Stephan Fuchs (Robert Koch Institute, MF1, fuchss@rki.de)

import os
import pickle
import sys

from Bio.Emboss.Applications import StretcherCommandline
import pandas as pd


class sonarAligner(object):
    """
    this object performs a pairwise sequence alignment and provides/stores selected
    alignment functionalities/statistics.
    """

    def __init__(self):
        self.nuc_profile = []
        self.nuc_n_profile = []
        self.aa_profile = []
        self.aa_n_profile = []

    def align(self, qry, ref, gapopen=16, gapextend=4):
        """ """
        cline = StretcherCommandline(
            asequence=qry,
            bsequence=ref,
            gapopen=gapopen,
            gapextend=gapextend,
            outfile="stdout",
            aformat="fasta",
        )
        stdout, stderr = cline()
        s1 = stdout.find("\n") + 1
        e = stdout[1:].find(">") + 1
        s2 = stdout[e:].find("\n") + e
        qry = stdout[s1:e].replace("\n", "")
        ref = stdout[s2:].replace("\n", "")
        return qry, ref

    def process_cached_sample(self, fname):
        """
        This function takes a sample file and processes it.
        create var file with NT and AA mutations
        """

        with open(fname, "rb") as handle:
            data = pickle.load(handle, encoding="bytes")

        if data["var_file"] is None:
            return True
        elif os.path.isfile(data["var_file"]):
            with open(data["var_file"], "r") as handle:
                for line in handle:
                    pass
            if line == "//":
                return True

        sourceid = str(data["sourceid"])
        alignment = self.align(data["seq_file"], data["ref_file"])
        nuc_vars = [x for x in self.extract_vars(*alignment, sourceid)]
        vars = "\n".join(["\t".join(x) for x in nuc_vars])
        if nuc_vars:
            # create AA mutation
            aa_vars = "\n".join(
                [
                    "\t".join(x)
                    for x in self.lift_vars(
                        nuc_vars, data["lift_file"], data["tt_file"]
                    )
                ]
            )
            if aa_vars:
                # concatinate to the same file of NT variants
                vars += "\n" + aa_vars
            vars += "\n"
        try:
            with open(data["var_file"], "w") as handle:
                handle.write(vars + "//")
        except OSError:
            os.makedirs(os.path.dirname(data["var_file"]), exist_ok=True)
            with open(data["var_file"], "w") as handle:
                handle.write(vars + "//")
        return True

    def extract_vars(self, qry_seq, ref_seq, elemid):
        query_length = len(qry_seq)
        if query_length != len(ref_seq):
            sys.exit("error: sequences differ in length")
        qry_seq += " "
        ref_seq += " "
        i = 0
        offset = 0
        while i < query_length:
            # match
            if qry_seq[i] == ref_seq[i]:
                pass
            # deletion
            elif qry_seq[i] == "-":
                s = i
                while qry_seq[i + 1] == "-":
                    i += 1
                start = s - offset
                end = i + 1 - offset
                if end - start == 1:
                    label = "del:" + str(start + 1)
                else:
                    label = "del:" + str(start + 1) + "-" + str(end)
                yield ref_seq[s : i + 1], str(start), str(end), " ", elemid, label

            # insertion
            elif ref_seq[i] == "-":
                s = i - 1
                while ref_seq[i + 1] == "-":
                    i += 1
                if s == -1:
                    ref = " "
                    alt = qry_seq[: i + 1]
                else:
                    ref = ref_seq[s]
                    alt = qry_seq[s : i + 1]
                pos = s - offset + 1
                yield ref, str(pos - 1), str(pos), alt, elemid, ref + str(pos) + alt
                offset += i - s
            # snps
            else:
                ref = ref_seq[i]
                alt = qry_seq[i]
                pos = i - offset + 1
                yield ref, str(pos - 1), str(pos), alt, elemid, ref + str(pos) + alt
            i += 1

    def translate(self, seq, tt):
        aa = []
        while len(seq) % 3 != 0:
            seq = seq[: len(seq) - 1]
        for codon in [seq[i : i + 3] for i in range(0, len(seq), 3)]:
            aa.append(tt[codon])
        return "".join(aa)

    def lift_vars(self, nuc_vars, lift_file, tt_file):
        df = pd.read_pickle(lift_file)
        with open(tt_file, "rb") as handle:
            tt = pickle.load(handle, encoding="bytes")
        for nuc_var in nuc_vars:
            for i in range(int(nuc_var[1]), int(nuc_var[2])):
                alt = "-" if nuc_var[3] == " " else nuc_var[3]
                df.loc[df["nucPos1"] == i, "alt1"] = alt
                df.loc[df["nucPos2"] == i, "alt2"] = alt
                df.loc[df["nucPos3"] == i, "alt3"] = alt

        df = df.loc[
            (df["ref1"] != df["alt1"])
            | (df["ref2"] != df["alt2"])
            | (df["ref3"] != df["alt3"])
        ]
        df["altAa"] = df.apply(
            lambda x: self.translate(x["alt1"] + x["alt1"] + x["alt1"], tt), axis=1
        )
        df = df.loc[df["aa"] != df["altAa"]]

        # snps or inserts
        for index, row in df.loc[(df["altAa"] != "-") & (df["altAa"] != "")].iterrows():
            pos = row["aaPos"] + 1
            label = row["aa"] + str(pos) + row["altAa"]
            yield row["aa"], str(pos - 1), str(pos), row["altAa"], str(
                row["elemid"]
            ), label

        # deletions
        prev_row = None
        for index, row in (
            df.loc[(df["altAa"] == "-")].sort_values(["elemid", "aaPos"]).iterrows()
        ):
            if prev_row is None:
                prev_row = row
            elif (
                prev_row["elemid"] == row["elemid"]
                and abs(prev_row["aaPos"] - row["aaPos"]) == 1
            ):
                prev_row["aa"] += row["aa"]
            else:
                start = prev_row["aaPos"]
                end = prev_row["aaPos"] + len(prev_row["aa"])
                if end - start == 1:
                    label = "del:" + str(start + 1)
                else:
                    label = "del:" + str(start + 1) + "-" + str(end)
                yield prev_row["aa"], str(start), str(end), " ", str(
                    prev_row["elemid"]
                ), label
                prev_row = row
        if prev_row is not None:
            start = prev_row["aaPos"]
            end = prev_row["aaPos"] + len(prev_row["aa"])
            if end - start == 1:
                label = "del:" + str(start + 1)
            else:
                label = "del:" + str(start + 1) + "-" + str(end)
            yield prev_row["aa"], str(start), str(end), " ", str(
                prev_row["elemid"]
            ), label



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# author: Stephan Fuchs (Robert Koch Institute, MF1, fuchss@rki.de)

import base64
from collections import defaultdict
import gzip
import hashlib
import lzma
import os
import pickle
from pickle import load as load_pickle
import pprint
import re
import shutil
import sys
from tempfile import mkdtemp
from tempfile import TemporaryDirectory

import pandas as pd
from tqdm import tqdm
import yaml

from .align import sonarAligner
from .basics import sonarBasics
from .dbm import sonarDBManager

pp = pprint.PrettyPrinter(indent=4)


class sonarCache:
    """ """

    def __init__(
        self,
        db=None,
        outdir=None,
        refacc=None,
        logfile=None,
        allow_updates=True,
        ignore_errors=False,
        temp=False,
        debug=False,
        disable_progress=False,
    ):
        self.db = db
        self.allow_updates = allow_updates
        self.debug = debug
        self.refacc = refacc
        self.temp = temp
        self.ignore_errors = ignore_errors
        self.disable_progress = disable_progress

        with sonarDBManager(self.db, debug=self.debug) as dbm:
            self.refmols = dbm.get_molecule_data(
                '"molecule.accession"',
                '"molecule.id"',
                '"molecule.standard"',
                '"translation.id"',
                reference_accession=self.refacc,
            )
            self.default_refmol_acc = [
                x for x in self.refmols if self.refmols[x]["molecule.standard"] == 1
            ][0]
            self.default_refmol_id = [
                x for x in self.refmols if self.refmols[x]["molecule.id"] == 1
            ][0]
            self.sources = {
                x["molecule.accession"]: dbm.get_source(x["molecule.id"])
                for x in self.refmols.values()
            }
            self.properties = dbm.properties
        self._propregex = re.compile(
            r"\[(" + "|".join(self.properties.keys()) + r")=([^\[\]=]+)\]"
        )
        self._molregex = re.compile(r"\[molecule=([^\[\]=]+)\]")

        self.logfile = open(logfile, "w") if logfile else None
        self.basedir = (
            os.path.abspath(mkdtemp(prefix=".sonarCache_"))
            if not outdir
            else os.path.abspath(outdir)
        )
        self.smk_config = os.path.join(self.basedir, "config.yaml")
        self.sample_dir = os.path.join(self.basedir, "samples")
        self.seq_dir = os.path.join(self.basedir, "seq")
        self.algn_dir = os.path.join(self.basedir, "algn")
        self.var_dir = os.path.join(self.basedir, "var")
        self.ref_dir = os.path.join(self.basedir, "ref")

        os.makedirs(self.basedir, exist_ok=True)
        os.makedirs(self.seq_dir, exist_ok=True)
        os.makedirs(self.ref_dir, exist_ok=True)
        # os.makedirs(self.algn_dir, exist_ok=True)
        os.makedirs(self.var_dir, exist_ok=True)
        os.makedirs(self.sample_dir, exist_ok=True)
        self._samplefiles = set()
        self._samplefiles_to_profile = set()
        self._refs = set()
        self._lifts = set()
        self._tt = set()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if os.path.isdir(self.basedir) and self.temp:
            shutil.rmtree(self.basedir)
        if self.logfile:
            self.logfile.close()
        if self.sample_yaml:
            self.sample_yaml.close()

    @staticmethod
    def slugify(string):
        return (
            base64.urlsafe_b64encode(string.encode("UTF-8")).decode("UTF-8").rstrip("=")
        )

    @staticmethod
    def deslugify(string):
        while len(string) % 3 != 0:
            string += "="
        return base64.urlsafe_b64decode(string).decode("utf-8")

    def log(self, msg, die=False, errtype="error"):
        if self.logfile:
            self.logfile.write(msg)
        elif not die:
            sys.stderr(msg)
        else:
            exit(errtype + ": " + msg)

    @staticmethod
    def write_pickle(fname, data):
        with open(fname, "wb") as handle:
            pickle.dump(data, handle)

    @staticmethod
    def read_pickle(fname):
        with open(fname, "rb") as handle:
            return pickle.load(handle, encoding="bytes")

    @staticmethod
    def pickle_collision(fname, data):
        if os.path.isfile(fname) and load_pickle(fname) != data:
            return True
        return False

    @staticmethod
    def file_collision(fname, data):
        with open(fname, "r") as handle:
            if handle.read() != data:
                return True
        return False

    def sample_collision(self, key, datadict):
        if sonarCache.read_pickle(key) != datadict:
            return True
        return False

    def write_smk_config(self):
        data = {
            "debug": self.debug,
            "sample_dir": self.sample_dir,
            "seq_dir": self.seq_dir,
            "algn_dir": self.algn_dir,
            "var_dir": self.var_dir,
        }

        with open(self.smk_config, "w") as handle:
            yaml.dump(data, handle)

    def get_seq_fname(self, seqhash):
        fn = self.slugify(seqhash)
        return os.path.join(self.seq_dir, fn[:2], fn + ".seq")

    def get_ref_fname(self, refid):
        return os.path.join(self.ref_dir, str(refid) + ".seq")

    def get_lift_fname(self, refid):
        return os.path.join(self.ref_dir, str(refid) + ".lift")

    def get_tt_fname(self, refid):
        return os.path.join(self.ref_dir, str(refid) + ".tt")

    def get_algn_fname(self, seqhash):
        fn = self.slugify(seqhash)
        return os.path.join(self.algn_dir, fn[:2], fn + ".algn")

    def get_var_fname(self, seqhash):
        fn = self.slugify(seqhash)
        return os.path.join(self.var_dir, fn[:2], fn + ".var")

    def get_sample_fname(self, sample_name):
        fn = self.slugify(hashlib.sha1(sample_name.encode("utf-8")).hexdigest())
        return os.path.join(self.sample_dir, fn[:2], fn + ".sample")

    def cache_sample(
        self,
        name,
        sampleid,
        seqhash,
        header,
        refmol,
        refmolid,
        sourceid,
        translation_id,
        algnid,
        seqfile,
        reffile,
        ttfile,
        algnfile,
        varfile,
        liftfile,
        properties,
    ):
        """
        The function takes in a bunch of arguments and returns a filename.
        :return: A list of dictionaries. Each dictionary contains the information for a single sample.
        """
        data = {
            "name": name,
            "sampleid": sampleid,
            "refmol": refmol,
            "refmolid": refmolid,
            "sourceid": sourceid,
            "translationid": translation_id,
            "algnid": algnid,
            "header": header,
            "seqhash": seqhash,
            "seq_file": seqfile,
            "ref_file": reffile,
            "tt_file": ttfile,
            "algn_file": algnfile,
            "var_file": varfile,
            "lift_file": liftfile,
            "properties": properties,
        }
        fname = self.get_sample_fname(name)  # full path
        try:
            self.write_pickle(fname, data)
        except OSError:
            os.makedirs(os.path.dirname(fname), exist_ok=True)
            self.write_pickle(fname, data)

        self._samplefiles.add(fname)
        if algnid is None:
            self._samplefiles_to_profile.add(fname)
        return fname

    def iter_samples(self):
        for fname in self._samplefiles:
            yield self.read_pickle(fname)

    def cache_sequence(self, seqhash, sequence):
        fname = self.get_seq_fname(seqhash)
        if os.path.isfile(fname):
            if self.file_collision(fname, sequence):
                sys.exit(
                    "seqhash collision: sequences differ for seqhash " + seqhash + "."
                )
        else:
            try:
                with open(fname, "w") as handle:
                    handle.write(sequence)
            except OSError:
                os.makedirs(os.path.dirname(fname), exist_ok=True)
                with open(fname, "w") as handle:
                    handle.write(sequence)
        return fname

    def cache_reference(self, refid, sequence):
        fname = self.get_ref_fname(refid)
        if refid not in self._refs:
            with open(fname, "w") as handle:
                handle.write(sequence)
            self._refs.add(refid)
        return fname

    def cache_translation_table(self, translation_id, dbm):
        """
        If the translation table
        is not in the cache, it is retrieved from the database and written to a file

        :param translation_id: The id of the translation table
        :param dbm: the database manager
        :return: A file name.
        """
        fname = self.get_tt_fname(translation_id)  # write under /cache/ref/
        if translation_id not in self._tt:
            self.write_pickle(fname, dbm.get_translation_dict(translation_id))
            self._tt.add(translation_id)
        return fname

    def cache_lift(self, refid, refmol_acc, sequence):
        """
                        The function takes in a reference id, a reference molecule accession number,
                        and a reference sequence. It then checks to see if the reference molecule accession number is in the set of molecules that
                        have been cached. If it is not, it iterates through all of the coding sequences for that molecule and creates a
                        dataframe for each one.
        .
                        It then saves the dataframe to a pickle file and adds the reference molecule accession number to
                        the set of molecules that have been cached.
                        It then returns the name of the pickle file

        """
        fname = self.get_lift_fname(refid)
        rows = []
        if refmol_acc not in self._lifts:
            cols = [
                "elemid",
                "nucPos1",
                "nucPos2",
                "nucPos3",
                "ref1",
                "ref2",
                "ref3",
                "alt1",
                "alt2",
                "alt3",
                "symbol",
                "aaPos",
                "aa",
            ]
            for cds in self.iter_cds(refmol_acc):
                elemid = cds["id"]
                symbol = cds["symbol"]
                seq = cds["sequence"] + "*"
                codon = 0
                i = 0
                coords = []
                for rng in cds["ranges"]:
                    coords.extend(list(rng))
                while len(coords) % 3 != 0:
                    coords.append("")
                coords_len = len(coords)
                while len(seq) < coords_len / 3:
                    seq += "-"
                while len(sequence) < coords_len:
                    sequence += "-"
                for i, coord in enumerate(
                    [coords[x : x + 3] for x in range(0, len(coords), 3)]
                ):
                    codon = [sequence[coord[0]], sequence[coord[1]], sequence[coord[2]]]
                    rows.append(
                        [elemid] + coord + codon + codon + [symbol, i, seq[i].strip()]
                    )
                df = pd.DataFrame.from_records(rows, columns=cols, coerce_float=False)
                df = df.reindex(df.columns.tolist(), axis=1)
                df.to_pickle(fname)
                if self.debug:
                    df.to_csv(fname + ".csv")
            self._lifts.add(refmol_acc)
        return fname

    @staticmethod
    def open_file(fname, mode="r", compressed="auto", encoding=None):
        if not os.path.isfile(fname):
            sys.exit("input error: " + fname + " does not exist.")
        if compressed == "auto":
            compressed = os.path.splitext(fname)[1][1:]
        try:
            if compressed == "gz":
                return gzip.open(fname, mode + "t", encoding=encoding)
            if compressed == "xz":
                return lzma.open(fname, mode + "t", encoding=encoding)
            return open(fname, mode, encoding=encoding)
        except Exception:
            sys.exit("input error: " + fname + " cannot be opened.")

    def process_fasta_entry(self, header, seq):
        sample_id = header.replace("\t", " ").replace("|", " ").split(" ")[0]
        refmol = self.get_refmol(header)
        if not refmol:
            sys.exit(
                "input error: "
                + sample_id
                + " refers to an unknown reference molecule ("
                + refmol
                + ")."
            )
        seq = sonarBasics.harmonize(seq)
        seqhash = sonarBasics.hash(seq)
        refmolid = self.refmols[refmol]["molecule.id"]
        return {
            "name": sample_id,
            "header": header,
            "seqhash": seqhash,
            "sequence": seq,
            "refmol": refmol,
            "refmolid": refmolid,
            "translation_id": self.refmols[refmol]["translation.id"],
            "properties": self.get_properties(header),
        }

    def iter_fasta(self, *fnames):
        """
        This function iterates over the fasta files and returns a dictionary for each record
        """
        for fname in fnames:
            with self.open_file(fname) as handle, tqdm(
                desc="processing " + fname + "...",
                total=os.path.getsize(fname),
                unit="bytes",
                unit_scale=True,
                bar_format="{desc} {percentage:3.0f}% [{n_fmt}/{total_fmt}, {elapsed}<{remaining}, {rate_fmt}{postfix}]",
                disable=self.disable_progress,
            ) as pbar:
                seq = []
                for line in handle:
                    pbar.update(len(line))
                    line = line.strip()
                    if line.startswith(">"):
                        header = line[1:]
                        if seq:
                            yield self.process_fasta_entry(header, "".join(seq))
                            seq = []
                    else:
                        seq.append(line)
                if seq:
                    yield self.process_fasta_entry(header, "".join(seq))

    def get_refmol(self, fasta_header):
        mol = self._molregex.search(fasta_header)
        if not mol:
            try:
                return self.refmols[mol]["accession"]
            except Exception:
                None
        return self.default_refmol_acc

    def get_refseq(self, refmol_acc):
        try:
            return self.sources[refmol_acc]["sequence"]
        except Exception:
            return None

    def iter_cds(self, refmol_acc):
        cds = {}
        prev_elem = None
        with sonarDBManager(self.db, debug=self.debug) as dbm:
            for row in dbm.get_annotation(
                molecule_accession=refmol_acc, element_type="cds"
            ):
                if prev_elem is None:
                    prev_elem = row["element.id"]
                elif row["element.id"] != prev_elem:
                    yield cds
                    cds = {}
                    prev_elem = row["element.id"]
                if cds == {}:
                    cds = {
                        "id": row["element.id"],
                        "accession": row["element.accession"],
                        "symbol": row["element.symbol"],
                        "sequence": row["element.sequence"],
                        "ranges": [
                            range(
                                row["elempart.start"],
                                row["elempart.end"],
                                row["elempart.strand"],
                            )
                        ],
                    }
                else:
                    cds["ranges"].append(
                        range(
                            row["elempart.start"],
                            row["elempart.end"],
                            row["elempart.strand"],
                        )
                    )
        if cds:
            yield cds

    def get_refseq_id(self, refmol_acc):
        try:
            return self.sources[refmol_acc]["id"]
        except Exception:
            return None

    def get_refhash(self, refmol_acc):
        try:
            if "seqhash" not in self.sources[refmol_acc]:
                self.sources[refmol_acc]["seqhash"] = sonarBasics.hash(
                    self.sources[refmol_acc]["sequence"]
                )
            return self.sources[refmol_acc]["seqhash"]
        except Exception:
            return None

    def get_properties(self, fasta_header):
        return {x.group(1): x.group(2) for x in self._propregex.finditer(fasta_header)}

    def add_fasta(self, *fnames, propdict=defaultdict(dict)):  # noqa: C901
        default_properties = {
            x: self.properties[x]["standard"] for x in self.properties
        }
        with sonarDBManager(self.db, debug=self.debug) as dbm:
            for fname in fnames:
                for data in self.iter_fasta(fname):
                    # check sample
                    data["sampleid"], seqhash = dbm.get_sample_data(data["name"])
                    data["sourceid"] = dbm.get_source(data["refmolid"])["id"]

                    # check properties
                    if data["sampleid"] is None:
                        props = default_properties
                        for k, v in data["properties"].items():
                            props[k] = v
                        for k, v in propdict[data["sampleid"]].items():
                            props[k] = v
                        data["properties"] = props
                    elif not self.allow_updates:
                        continue
                    else:
                        for k, v in propdict[data["sampleid"]].items():
                            data["properties"][k] = v

                    # check reference
                    refseq_id = self.get_refseq_id(data["refmol"])
                    if not refseq_id:
                        if not self.ignore_errors:
                            self.log(
                                "fasta header refers to an unknown refrence ("
                                + data["header"]
                                + ")",
                                True,
                                "input error",
                            )
                        else:
                            self.log(
                                "skipping "
                                + data["name"]
                                + " referring to an unknown reference ("
                                + data["header"]
                                + ")"
                            )

                    # check alignment
                    data["algnid"] = dbm.get_alignment_id(data["seqhash"], refseq_id)
                    if data["algnid"] is None:
                        data["seqfile"] = self.cache_sequence(
                            data["seqhash"], data["sequence"]
                        )
                        data["reffile"] = self.cache_reference(
                            refseq_id, self.get_refseq(data["refmol"])
                        )
                        data["ttfile"] = self.cache_translation_table(
                            data["translation_id"], dbm
                        )
                        data["liftfile"] = self.cache_lift(
                            refseq_id, data["refmol"], self.get_refseq(data["refmol"])
                        )
                        data["algnfile"] = self.get_algn_fname(
                            data["seqhash"] + "@" + self.get_refhash(data["refmol"])
                        )
                        data["varfile"] = self.get_var_fname(
                            data["seqhash"] + "@" + self.get_refhash(data["refmol"])
                        )
                    elif data["seqhash"] != seqhash:
                        data["seqfile"] = self.cache_sequence(
                            data["seqhash"], data["sequence"]
                        )
                        data["reffile"] = None
                        data["ttfile"] = None
                        data["liftfile"] = None
                        data["algnfile"] = None
                        data["varfile"] = None
                    else:
                        data["seqhash"] = None
                        data["seqfile"] = None
                        data["reffile"] = None
                        data["ttfile"] = None
                        data["liftfile"] = None
                        data["algnfile"] = None
                        data["varfile"] = None
                    del data["sequence"]
                    self.cache_sample(**data)

    def import_cached_samples(self):  # noqa: C901
        refseqs = {}
        with sonarDBManager(self.db, readonly=False, debug=self.debug) as dbm:
            for sample_data in tqdm(
                self.iter_samples(),
                total=len(self._samplefiles),
                desc="importing samples...",
                unit="samples",
                bar_format="{desc} {percentage:3.0f}% [{n_fmt}/{total_fmt}, {elapsed}<{remaining}, {rate_fmt}{postfix}]",
                disable=self.disable_progress,
            ):
                try:
                    # nucleotide level import
                    if not sample_data["seqhash"] is None:
                        # sampleid = dbm.insert_sample(
                        #     sample_data["name"], sample_data["seqhash"]
                        # )
                        algnid = dbm.insert_alignment(
                            sample_data["seqhash"], sample_data["refmolid"]
                        )
                    if not sample_data["var_file"] is None:
                        with open(sample_data["var_file"], "r") as handle:
                            for line in handle:
                                if line == "//":
                                    break
                                vardat = line.strip("\r\n").split("\t")
                                dbm.insert_variant(
                                    algnid,
                                    vardat[4],
                                    vardat[0],
                                    vardat[3],
                                    vardat[1],
                                    vardat[2],
                                    vardat[5],
                                )
                            if line != "//":
                                sys.exit(
                                    "cache error: corrupted file ("
                                    + sample_data["var_file"]
                                    + ")"
                                )
                    # paranoia test
                    if not sample_data["seqhash"] is None:
                        try:
                            seq = list(refseqs[sample_data["refmolid"]])
                        except Exception:
                            refseqs[sample_data["refmolid"]] = list(
                                dbm.get_sequence(sample_data["refmolid"])
                            )
                            seq = list(refseqs[sample_data["refmolid"]])

                        prefix = ""
                        for vardata in dbm.iter_dna_variants(
                            sample_data["name"], sample_data["refmolid"]
                        ):
                            if vardata["variant.alt"] == " ":
                                for i in range(
                                    vardata["variant.start"], vardata["variant.end"]
                                ):
                                    seq[i] = ""
                            elif vardata["variant.start"] >= 0:
                                seq[vardata["variant.start"]] = vardata["variant.alt"]
                            else:
                                prefix = vardata["variant.alt"]
                        seq = prefix + "".join(seq)

                        with open(sample_data["seq_file"], "r") as handle:
                            orig_seq = handle.read()

                        if seq != orig_seq:
                            aligner = sonarAligner()
                            with TemporaryDirectory() as tempdir:
                                qryfile = os.path.join(tempdir, "qry")
                                reffile = os.path.join(tempdir, "ref")
                                with open(qryfile, "w") as handle:
                                    handle.write(">seq\n" + seq)
                                with open(reffile, "w") as handle:
                                    handle.write(">ref\n" + orig_seq)
                                qry, ref = aligner.align(qryfile, reffile)
                            with open("paranoid.alignment.fna", "w") as handle:
                                handle.write(
                                    ">original_"
                                    + sample_data["name"]
                                    + "\n"
                                    + ref
                                    + "\n>restored_"
                                    + sample_data["name"]
                                    + "\n"
                                    + qry
                                )
                            sys.exit(
                                "import error: original sequence of sample "
                                + sample_data["name"]
                                + " cannot be restored from stored genomic profile for sample (see paranoid.alignment.fna)"
                            )

                except Exception as e:
                    print("\n------- Fatal Error ---------")
                    print("\nDebugging Information")
                    print(e)
                    pp.pprint(sample_data)
                    sys.exit("unknown import error")


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# author: Stephan Fuchs (Robert Koch Institute, MF1, fuchss@rki.de)

import argparse
from collections import defaultdict
import os
import sys
from tempfile import mkstemp
from textwrap import fill

from mpire import WorkerPool
from tabulate import tabulate
from tqdm import tqdm

from . import sonardb
from .align import sonarAligner
from .basics import sonarBasics
from .cache import sonarCache
from .dbm import sonarDBManager
from .linmgr import sonarLinmgr


class arg_namespace(object):
    pass


def parse_args(args):
    """
    setting and handling command line arguments
    """

    VERSION = sonarBasics.get_version()

    # preparations
    user_namespace = arg_namespace()
    parser = argparse.ArgumentParser(prog="sonar", description="covsonar " + VERSION)
    subparsers = parser.add_subparsers(
        help="detect, store, and screen for mutations in genomic sequences"
    )
    subparsers.dest = "tool"
    subparsers.required = True

    # parser components
    # parser component: db input
    general_parser = argparse.ArgumentParser(add_help=False)
    general_parser.add_argument(
        "--db", metavar="FILE", help="sonar database to use", type=str, required=True
    )
    general_parser.add_argument(
        "--debug",
        help="activate debugging mode showing all sqllite queries on screen",
        action="store_true",
    )

    # parser component: output
    output_parser = argparse.ArgumentParser(add_help=False)
    output_parser.add_argument(
        "-o",
        "--out",
        metavar="FILE",
        help="write output to file",
        type=str,
        default=None,
    )

    # parser component: sample input
    sample_parser = argparse.ArgumentParser(add_help=False)
    sample_parser.add_argument(
        "--sample",
        metavar="STR",
        help="sample accession(s) to consider ",
        type=str,
        nargs="+",
        default=[],
    )
    sample_parser.add_argument(
        "--sample-file",
        metavar="FILE",
        help="file containing sample accession(s) to consider (one per line)",
        type=str,
        nargs="+",
        default=[],
    )

    # parser component: property name
    prop_parser = argparse.ArgumentParser(add_help=False)
    prop_parser.add_argument(
        "--name", metavar="STR", help="name of sample property", type=str, required=True
    )

    # parser component: reference
    ref_parser = argparse.ArgumentParser(add_help=False)
    ref_parser.add_argument(
        "-r",
        "--reference",
        metavar="STR",
        help="reference accession",
        type=str,
        default=None,
    )

    # parser component: threads
    thread_parser = argparse.ArgumentParser(add_help=False)
    thread_parser.add_argument(
        "-t",
        "--threads",
        metavar="INT",
        help="number of threads to use (default: 1)",
        type=int,
        default=1,
    )

    # main parser
    # setup parser
    parser_setup = subparsers.add_parser(
        "setup", parents=[general_parser], help="Setup a new database."
    )
    parser_setup.add_argument(
        "-a",
        "--auto-create",
        help="Auto create important properties",
        action="store_true",
    )
    parser_setup.add_argument(
        "--gbk",
        metavar="FILE",
        help="genbank file of the reference genome (default MN908947.3 is used as reference)",
        type=str,
        default=None,
    )

    # import parser
    parser_import = subparsers.add_parser(
        "import",
        parents=[general_parser, thread_parser],
        help="Import genome sequences and sample information to the database.",
    )
    parser_import.add_argument(
        "--fasta",
        help="fasta file containing genome sequences to import",
        type=str,
        nargs="+",
        default=None,
    )
    parser_import.add_argument(
        "--tsv",
        help="tab-delimited file containing sample properties to import",
        type=str,
        nargs="+",
        default=None,
    )
    parser_import.add_argument(
        "--cols",
        help="define column names for sample properties (if different from property name)",
        type=str,
        nargs="+",
        default=[],
    )
    parser_import.add_argument(
        "--no-autodetect",
        help="do not auto-detect of columns for sample properties based on property names",
        action="store_true",
    )
    parser_import.add_argument(
        "--no-update",
        help="skip samples already existing in the database",
        action="store_true",
    )
    parser_import.add_argument(
        "--cache",
        metavar="DIR",
        help="directory for chaching data (default: a temporary directory is created)",
        type=str,
        default=None,
    )
    parser_import_g1 = parser.add_mutually_exclusive_group()
    parser_import_g1.add_argument(
        "--no-progress",
        "-p",
        help="don't show progress bars while importing",
        action="store_true",
    )

    # view-prop parser
    subparsers.add_parser(
        "list-prop",
        parents=[general_parser],
        help="View sample properties added to the database.",
    )

    # add-prop parser
    parser_addprops = subparsers.add_parser(
        "add-prop",
        parents=[general_parser, prop_parser],
        help="Add a sample property to the database.",
    )
    parser_addprops.add_argument(
        "--descr",
        metavar="STR",
        help="description of the new property",
        type=str,
        required=True,
    )
    parser_addprops.add_argument(
        "--dtype",
        metavar="STR",
        help="data type of the new property",
        type=str,
        choices=["integer", "float", "text", "date", "zip"],
        required=True,
    )
    parser_addprops.add_argument(
        "--qtype",
        metavar="STR",
        help="query type of the new property",
        type=str,
        choices=["numeric", "float", "date", "text", "zip"],
        default=None,
    )
    parser_addprops.add_argument(
        "--default",
        metavar="VAR",
        help="default value of the new property (none by default)",
        type=str,
        default=None,
    )

    # delete-prop parser
    parser_delprops = subparsers.add_parser(
        "delete-prop",
        parents=[general_parser, prop_parser],
        help="Delete a sample property from the database.",
    )
    parser_delprops.add_argument(
        "--force", help="force property to be deleted", action="store_true"
    )

    # match parser
    parser_match = subparsers.add_parser(
        "match",
        parents=[sample_parser, output_parser, general_parser],
        help="get mutations profiles for given accessions.",
    )
    parser_match.add_argument(
        "--profile",
        "-p",
        metavar="STR",
        help="match genomes sharing the given mutation profile",
        type=str,
        action="append",
        nargs="+",
        default=[],
    )
    parser_match_format = parser_match.add_mutually_exclusive_group()
    parser_match_format.add_argument(
        "--count", help="count instead of listing matching genomes", action="store_true"
    )
    parser_match_format.add_argument(
        "--format",
        help="output format (default: tsv)",
        choices=["csv", "tsv", "vcf"],
        default="tsv",
    )
    parser_match.add_argument(
        "--with-sublineage",
        metavar="STR",
        help="recursively get all sublineages from a given lineage. ",
        type=str,
        default=None,
    )

    # delete parser
    parser_delete = subparsers.add_parser(
        "delete",
        parents=[ref_parser, sample_parser, general_parser],
        help="delete one or more samples from the database.",
    )
    parser_delete.add_argument(
        "--aligned",
        help="ise aligned form (deletions indicated by - and insertions by lower-case letters)",
        action="store_true",
    )

    # restore parser
    parser_restore = subparsers.add_parser(
        "restore",
        parents=[ref_parser, sample_parser, general_parser],
        help="restore sequence(s) from the database.",
    )
    parser_restore.add_argument(
        "--aligned",
        help="ise aligned form (deletions indicated by - and insertions by lower-case letters)",
        action="store_true",
    )

    # info parser
    subparsers.add_parser(
        "info", parents=[general_parser], help="show software and database info"
    )

    # optimize parser
    subparsers.add_parser(
        "optimize", parents=[general_parser], help="optimizes the database."
    )

    # dev parser
    subparsers.add_parser("dev", parents=[general_parser])

    # db-upgrade parser
    subparsers.add_parser(
        "db-upgrade",
        parents=[general_parser],
        help="upgrade a database to the latest version",
    )

    # update-lineage-info parser
    subparsers.add_parser(
        "update-lineage-info",
        parents=[general_parser],
        help="download latest lineage information",
    )

    # version parser
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="covsonar " + VERSION,
        help="Show program's version number and exit.",
    )

    # register known arguments
    pargs = parser.parse_known_args(args=args, namespace=user_namespace)

    # register dynamic arguments
    # register additonal arguments from database-specific sample property names when matching genomes
    if pargs[0].tool == "match" and hasattr(pargs[0], "db"):
        with sonarDBManager(pargs[0].db, readonly=True, debug=pargs[0].debug) as dbm:
            for prop in dbm.properties.values():
                if prop["datatype"] == "integer":
                    t = int
                elif prop["datatype"] == "float":
                    t = float
                else:
                    t = str
                if pargs[0].tool == "match":
                    parser_match.add_argument(
                        "--" + prop["name"],
                        type=t,
                        nargs="+",
                        default=argparse.SUPPRESS,
                    )

    # return
    return parser.parse_args(args=args, namespace=user_namespace)


class sonar:
    def __init__(self, db, gff=None, debug=False):
        self.dbfile = db if db else mkstemp()[1]
        self.db = sonardb.sonarDB(self.dbfile)
        self.gff = gff
        self.debug = debug

    def show_system_info(self):
        """
        Prints out the version of the database, the name of the reference genome, the length of the
        reference genome, the names of the annotated proteins, and the translation table used
        """
        print("sonarDB version:       ", self.db.get_version())
        print("reference genome:      ", self.db.refdescr)
        print("reference length:      ", str(len(self.db.refseq)) + "bp")
        print("annotated proteins:    ", ", ".join(self.db.refgffObj.symbols))
        print("used translation table:", self.db.translation_table)

    def show_db_info(self):
        with sonardb.sonarDBManager(self.dbfile, readonly=True) as dbm:
            print("database path:             ", dbm.dbfile)
            print("database version:          ", dbm.get_db_version())
            print("database size:             ", self.get_db_size())
            g = dbm.count_genomes()
            print("genomes:                   ", g)
            print("unique sequences:          ", dbm.count_sequences())
            print("labs:                      ", dbm.count_labs())
            print("earliest genome import:    ", dbm.get_earliest_import())
            print("latest genome import:      ", dbm.get_latest_import())
            print("earliest sampling date:    ", dbm.get_earliest_date())
            print("latest sampling date:      ", dbm.get_latest_date())
            print("metadata:          ")
            fields = sorted(
                [
                    "lab",
                    "source",
                    "collection",
                    "technology",
                    "platform",
                    "chemistry",
                    "software",
                    "software_version",
                    "material",
                    "ct",
                    "gisaid",
                    "ena",
                    "lineage",
                    "zip",
                    "date",
                ]
            )
            maxlen = max([len(x) for x in fields])
            for field in fields:
                if g == 0:
                    c = 0
                    p = 0
                else:
                    c = dbm.count_metadata(field)
                    p = c / g * 100
                spacer = " " * (maxlen - len(field))
                print("   " + field + " information:" + spacer, f"{c} ({p:.{2}f}%)")


def check_file(fname):
    if not os.path.isfile(fname):
        sys.exit("iput error: " + fname + " is not a valid file.")


def main(args):  # noqa: C901
    # process arguments
    # args = parse_args()

    if hasattr(args, "db") and args.db:
        if args.tool != "setup" and args.db is not None and not os.path.isfile(args.db):
            sys.exit("input error: database does not exist.")
            # check_db_compatibility

    # set debugging mode
    if hasattr(args, "debug") and args.debug:
        debug = True
    else:
        debug = False

    # tool procedures
    # setup, db-upgrade
    if args.tool == "setup":
        if args.gbk:
            check_file(args.gbk)
        sonarBasics.setup_db(
            args.db, args.auto_create, reference_gb=args.gbk, debug=debug
        )
    elif args.tool == "db-upgrade":
        input("Warning: Backup db file before upgrading, Press Enter to continue...")
        sonarDBManager.upgrade_db(args.db)
    else:
        with sonarDBManager(args.db, readonly=True) as dbm:
            dbm.check_db_compatibility()
    # other than the above
    # import
    if args.tool == "import":
        if args.no_update:
            update = False
            print("import mode: skipping existing samples")
            if args.fasta is None:
                print("Nothing to import.")
                exit(0)
        else:
            update = True
            print("import mode: updating existing samples")

        if args.tsv is None and args.fasta is None:
            print("Nothing to import.")
            exit(0)

        # prop handling
        with sonarDBManager(args.db, readonly=True) as dbm:
            db_properties = set(dbm.properties.keys())
            db_properties.add("sample")

        colnames = {} if args.no_autodetect else {x: x for x in db_properties}
        for x in args.cols:
            if x.count("=") != 1:
                sys.exit(
                    "input error: " + x + " is not a valid sample property assignment."
                )
            k, v = x.split("=")
            if k not in db_properties:
                sys.exit(
                    "input error: sample property "
                    + k
                    + " is unknown to the selected database. Use list-props to see all valid properties."
                )
            colnames[k] = v

        if "sample" not in colnames:
            sys.exit("input error: a sample column has to be assigned.")

        properties = defaultdict(dict)
        for tsv in args.tsv:
            with open(tsv, "r") as handle, tqdm(
                desc="processing " + tsv + "...",
                total=os.path.getsize(tsv),
                unit="bytes",
                unit_scale=True,
                bar_format="{desc} {percentage:3.0f}% [{n_fmt}/{total_fmt}, {elapsed}<{remaining}, {rate_fmt}{postfix}]",
                disable=args.no_progress,
            ) as pbar:
                line = handle.readline()
                pbar.update(len(line))
                fields = line.strip("\r\n").split("\t")
                tsv_cols = {}
                print()
                for x in sorted(colnames.keys()):
                    c = fields.count(colnames[x])
                    if c == 1:
                        tsv_cols[x] = fields.index(colnames[x])
                        print("  " + x + " <- " + colnames[x])
                    elif c > 1:
                        sys.exit("error: " + colnames[x] + " is not an unique column.")
                if "sample" not in tsv_cols:
                    sys.exit("error: tsv file does not contain required sample column.")
                elif len(tsv_cols) == 1:
                    sys.exit(
                        "input error: tsv does not provide any informative column."
                    )
                for line in handle:
                    pbar.update(len(line))
                    fields = line.strip("\r\n").split("\t")
                    sample = fields[tsv_cols["sample"]]
                    for x in tsv_cols:
                        if x == "sample":
                            continue
                        properties[sample][x] = fields[tsv_cols[x]]

        # setup cache
        temp = False if args.cache else True
        cache = sonarCache(
            args.db,
            outdir=args.cache,
            logfile="import.log",
            allow_updates=update,
            temp=temp,
            debug=args.debug,
            disable_progress=args.no_progress,
        )

        # importing sequences
        if args.fasta is not None:
            cache.add_fasta(*args.fasta, propdict=properties)

            aligner = sonarAligner()
            l = len(cache._samplefiles_to_profile)
            with WorkerPool(n_jobs=args.threads, start_method="fork") as pool, tqdm(
                desc="profiling sequences...",
                total=l,
                unit="seqs",
                bar_format="{desc} {percentage:3.0f}% [{n_fmt}/{total_fmt}, {elapsed}<{remaining}, {rate_fmt}{postfix}]",
                disable=args.no_progress,
            ) as pbar:
                for _ in pool.imap_unordered(
                    aligner.process_cached_sample, cache._samplefiles_to_profile
                ):
                    pbar.update(1)

            cache.import_cached_samples()

        # importing properties
        if args.tsv:
            with sonarDBManager(args.db, readonly=False, debug=args.debug) as dbm:
                for sample_name in tqdm(
                    properties,
                    desc="import data ...",
                    total=len(properties),
                    unit="samples",
                    bar_format="{desc} {percentage:3.0f}% [{n_fmt}/{total_fmt}, {elapsed}<{remaining}, {rate_fmt}{postfix}]",
                    disable=args.no_progress,
                ):
                    sample_id = dbm.get_sample_id(sample_name)
                    if not sample_id:
                        continue
                    for property_name, value in properties[sample_name].items():
                        dbm.insert_property(sample_id, property_name, value)

    # view-prop
    elif args.tool == "list-prop":
        with sonarDBManager(args.db, debug=debug) as dbm:
            if not dbm.properties:
                print("*** no properties ***")
                exit(0)

            cols = [
                "name",
                "argument",
                "description",
                "data type",
                "query type",
                "standard value",
            ]
            rows = []
            for prop in sorted(dbm.properties.keys()):
                dt = (
                    dbm.properties[prop]["datatype"]
                    if dbm.properties[prop]["datatype"] != "float"
                    else "decimal number"
                )
                rows.append([])
                rows[-1].append(prop)
                rows[-1].append("--" + prop)
                rows[-1].append(fill(dbm.properties[prop]["description"], width=25))
                rows[-1].append(dt)
                rows[-1].append(dbm.properties[prop]["querytype"])
                rows[-1].append(dbm.properties[prop]["standard"])

            print(tabulate(rows, headers=cols, tablefmt="fancy_grid"))
            print()
            print("DATE FORMAT")
            print("dates must comply with the following format: YYYY-MM-DD")
            print()
            print("OPERATORS")
            print(
                "integer, floating point and decimal number data types support the following operators prefixed directly to the respective value without spaces:"
            )
            print("  > larger than (e.g. >1)")
            print("  < smaller than (e.g. <1)")
            print("  >= larger than or equal to (e.g. >=1)")
            print("  <= smaller than or equal to (e.g. <=1)")
            print("  != different than (e.g. !=1)")
            print()
            print("RANGES")
            print(
                "integer, floating point and date data types support ranges defined by two values directly connected by a colon (:) with no space between them:"
            )
            print("  e.g. 1:10 (between 1 and 10)")
            print("  e.g. 2021-01-01:2021-12-31 (between 1st Jan and 31st Dec of 2021)")
            print()

    # add-prop
    elif args.tool == "add-prop":
        with sonarDBManager(args.db, readonly=False, debug=args.debug) as dbm:
            if args.qtype is None:
                if args.dtype == "integer":
                    args.qtype = "numeric"
                elif args.dtype == "float":
                    args.qtype = "numeric"
                elif args.dtype == "text":
                    args.qtype = "text"
                elif args.dtype == "date":
                    args.qtype = "date"
                elif args.dtype == "zip":
                    args.qtype = "zip"
            dbm.add_property(
                args.name, args.dtype, args.qtype, args.descr, args.default
            )
        print("Inserted successfully:", args.name)

    # delprop
    elif args.tool == "delete-prop":
        with sonarDBManager(args.db, readonly=False, debug=debug) as dbm:
            if args.name not in dbm.properties:
                sys.exit("input error: unknown property.")
            a = dbm.count_property(args.name)
            b = dbm.count_property(args.name, ignore_standard=True)
            if args.force:
                decision = "YES"
            else:
                print(
                    "WARNING: There are",
                    a,
                    "samples with content for this property. Amongst those,",
                    b,
                    "samples do not share the default value of this property.",
                )
                decision = ""
                while decision not in ("YES", "no"):
                    decision = input(
                        "Do you really want to delete this property? [YES/no]: "
                    )
            if decision == "YES":
                dbm.delete_property(args.name)
                print("property deleted.")
            else:
                print("property not deleted.")

    # delete
    elif args.tool == "delete":
        samples = set([x.strip() for x in args.sample])
        for file in args.sample_file:
            check_file(file)
            with sonarBasics.open_file(file, compressed="auto") as handle:
                for line in handle:
                    samples.add(line.strip())
        if len(samples) == 0:
            print("Nothing to delete.")
        else:
            sonarBasics.delete(args.db, *samples, debug=args.debug)

    # restore
    elif args.tool == "restore":
        samples = set([x.strip() for x in args.sample])
        for file in args.sample_file:
            check_file(file)
            with sonarBasics.open_file(file, compressed="auto") as handle:
                for line in handle:
                    samples.add(line.strip())
        if len(samples) == 0:
            print("Nothing to restore.")
        else:
            sonarBasics.restore(
                args.db, *samples, aligned=args.aligned, debug=args.debug
            )

    # update-lineage-info
    elif args.tool == "update-lineage-info":
        print("Start to update parent-child relationship")
        fname = os.path.join(
            os.path.join(os.path.dirname(os.path.realpath(__file__))),
            "data/lineage.all.tsv",
        )
        # obj = sonarLinmgr()
        # lin_df = obj.update_lineage_data(fname)
        with sonarLinmgr() as obj:
            lin_df = obj.update_lineage_data(fname)

        with sonarDBManager(args.db, readonly=False, debug=args.debug) as dbm:
            dbm.add_update_lineage(lin_df)
            print("Update has been successfully")

    # info
    elif args.tool == "info":
        sonarBasics.show_db_info(args.db)

    # match
    elif args.tool == "match":
        props = {}
        reserved_props = {}

        with sonarDBManager(args.db, readonly=False, debug=args.debug) as dbm:
            for pname in dbm.properties:
                if hasattr(args, pname):
                    props[pname] = getattr(args, pname)
            if args.with_sublineage:
                if args.with_sublineage in dbm.properties:
                    reserved_props["with_sublineage"] = args.with_sublineage
                else:
                    sys.exit(
                        "input error: with-sublineage value is mismatch to the available properties"
                    )

        # for reserved keywords
        reserved_key = ["sample"]
        for pname in reserved_key:
            if hasattr(args, pname):
                if pname == "sample" and len(getattr(args, pname)) > 0:
                    # reserved_props[pname] = set([x.strip() for x in args.sample])
                    reserved_props = sonarBasics.set_key(
                        reserved_props, pname, getattr(args, pname)
                    )
                    # reserved_props[pname] = getattr(args, pname)

        # Support file upload
        if args.sample_file:
            for sample_file in args.sample_file:
                check_file(sample_file)
                with sonarBasics.open_file(sample_file, compressed="auto") as file:
                    for line in file:
                        reserved_props = sonarBasics.set_key(
                            reserved_props, "sample", line.strip()
                        )
        format = "count" if args.count else args.format

        sonarBasics.match(
            args.db,
            args.profile,
            reserved_props,
            props,
            outfile=args.out,
            debug=args.debug,
            format=format,
        )

    # optimize
    if args.tool == "optimize":
        with sonarDBManager(args.db, debug=args.debug) as dbm:
            dbm.optimize(args.db)

    # dev
    if args.tool == "dev":
        print("***dev mode***")
        with sonarDBManager(args.db, debug=debug) as dbm:
            for feature in dbm.get_annotation():
                print(())
    # Finished successfully
    return 0


def run():
    parsed_args = parse_args(sys.argv[1:])
    main(parsed_args)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['covsonar']

package_data = \
{'': ['*'], 'covsonar': ['data/*']}

install_requires = \
['biopython>=1.79,<1.80',
 'more-itertools>=8.7.0,<8.8.0',
 'mpire>=2.3.3,<2.4.0',
 'pandas>=1.4.0,<1.5.0',
 'pyaml>=20.4.0,<20.5.0',
 'requests>=2.28.0,<3.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'tqdm>=4.59.0,<4.60.0']

entry_points = \
{'console_scripts': ['sonar = covsonar.sonar:run']}

setup_kwargs = {
    'name': 'covsonar',
    'version': '2.0.0a0',
    'description': 'A database-driven system for handling genomic sequences and screening genomic profiles.',
    'long_description': '<img src="logo.png"  width="134" height="134" align="right"><br><br><br><br><br>\n\n# covSonar2\n\ncovSonar is a database-driven system for handling genomic sequences and screening genomic profiles.\n\nWhat\'s new in covSonar V.2\n* New design\n    * Improve workflows\n    * Performance improvements\n* Exciting new features\n\t* Support multiple pathogens\n\t* Flexible in adding meta information \n* New database design\n\t* New database schema \n    * Retrieval efficiency\n\t* Significantly smaller than the previous version\n\n## 1. Prerequisites / Setup\n\ncovSonar2 has some software-environmental requirements that can most easily be met by building a custom conda environment.\n\nProceed as follows to install covSonar:\n```sh\n# download the repository to the current working directory using git \ngit clone https://github.com/rki-mf1/covsonar\n# build the custom software environment using conda [recommended]\nconda env create -n sonar2 -f covsonar/sonar.env.yml\n# activate the conda evironment if built \nconda activate sonar2\n# testing\n./covsonar/test.sh\n```\n\n## 3. Usage\n\nIn covSonar2 there are several tools that can be called via subcommands.  \n\n| subcommand | purpose                                                             |\n|------------|---------------------------------------------------------------------|\n| setup      | setup a new database.                             |\n| import     | import genome sequences and sample information to the database     |\n| list-prop  | view sample properties added to the database             | \n| add-prop    | add a sample property to the database                    |\n| delete-prop       | delete a sample property from the database |\n| match   |  get mutations profiles for given accessions                                        |\n| restore   | restore sequence(s) from the database                                         |\n| info   |  show software and database info.                                        |\n| optimize   | optimizes the database                                         |\n| db-upgrade   | upgrade a database to the latest version                                         |\n| update-lineage-info | download latest lineage information                                       |\n\nEach tool provides a help page that can be accessed with the `-h` option.\n\n```sh\n# display help page\n./sonar.py -h \n# display help page for each tool\n./sonar.py import -h \n```\n\n### 3.1 Setup database (setup ⛽)\nFirst, we have to create a new database instance.\n```sh\n./sonar.py setup --db test.db\n```\n\n, or we can create a new database instance with predefined properties.\n```sh\n./sonar.py setup --db test.db --auto-create\n```\n\nBy default, the MN908947.3 (SARS-CoV-2) is used as a reference. If we want to set up database for a different pathogen, we can add `--gbk` following with Genbank file.\n\nexample;\n```sh\n./sonar.py setup --db test.db --auto-create --gbk Ebola.gb\n```\n\n>  note 📌:  [how to download genbank file](https://ncbiinsights.ncbi.nlm.nih.gov/2017/05/08/genome-data-download-made-easy/)\n\n### 3.2 Property management (list-prop, add-prop, delete-prop)\n\nIn covSonar2, users now can arbitrarily add meta information or properties into a database to fit a specific project objective.s\n\nTo view the added properties, we can use the `list-prop` command to display all information.\n```sh\n./sonar.py list-prop --db test.db \n```\n\nTo add more properties, we can use the `add-prop` command to add meta information into the database.\n\nThe required arguments are listed below when we use `add-prop`\n* `--name`, name of sample property\n* `--descr`, description of the new property\n* `--dtype`, data type of the new property (e.g., \'integer\', \'float\', \'text\', \'date\', \'zip\')\n\n```sh\n./sonar.py add-prop --db test.db  --name SEQ_REASON --dtype text --descr "seq. reason"\n```\n> tip 🕯️:  `./sonar.py add-prop -h ` to see all available   arguments\n\n\nThe `delete-prop` command is used to delete an unwanted property from the database.\n\n```sh\n./sonar.py delete-prop --db test.db  --name SEQ_REASON\n```\n\nThe program will ask an user for confirmation of the action.\n```\nDo you really want to delete this property? [YES/no]: YES\n```\n\n### 3.3 Adding genomes and meta information to the database (import)\n\nAdd sequence with meta information \n\n```sh\n./sonar.py import --db test.db --fasta valid.fasta --tsv day.tsv --threads 64 --cache tmp_cache --cols sample=IMS_ID\n```\n\nUpdate more\n\nexample:\n```sh\n\n./sonar.py import --db test.db --fasta valid.fasta --tsv day.tsv --threads 64 --cache tmp_cache --cols sample=IMS_ID\n\n./sonar.py import --db test.db --fasta valid.fasta --tsv day.tsv --threads 64 --cache tmp_cache --cols sample=IMS_ID\n```\n\n\n`sample` \n\n### 3.4 Query genome sequences based on profiles (match)\n\nGenomic profiles can be defined to align genomes. For this purpose, the variants related to the complete genome of the SARS-CoV-2 isolate Wuhan-Hu-1 (NC_045512.2) must be expressed as follows:\n\n| type       | nucleotide level                                                  | amino acid level              |\n|-----------|-------------------------------------------------------------------|-------------------------------|\n| SNP       | ref_nuc _followed by_ ref_pos _followed by_ alt_nuc (e.g. A3451T) | protein_symbol:ref_aa _followed by_ ref_pos _followed by_ alt_aa (e.g. S:N501Y) |\n| deletion  | del:ref_pos:length_in_bp (e.g. del:3001:8)                        | protein_symbol:del:ref_pos:length_in_aa (e.g. ORF1ab:del:3001:21) | \n| insertion | ref_nuc _followed by_ ref_pos _followed by_ alt_nucs (e.g. A3451TGAT) | protein_symbol:ref_aa _followed by_ ref_pos _followed by_ alt_aas (e.g. N:A34AK)  |  \n\nThe positions refer to the reference (first nucleotide in the genome is position 1). Using the option `--profile`, multiple variant definitions can be combined into a nucleotide, amino acid or mixed profile, which means that matching genomes must have all those variations in common. In contrast, alternative variations can be defined by multiple `--profile` options. As an example, `--profile S:N501Y S:E484K` matches genomes sharing the _Nelly_ **AND** _Erik_ variation while `--profile S:N501Y --profile S:E484K` matches to genomes that share either the _Nelly_ **OR** _Erik_ variation **OR** both. Accordingly, using the option **^** profiles can be defined that have not to be present in the matched genomes. \n\n\nThere are additional options to adjust the matching.\n\n| option             | description                                                            |\n|--------------------|------------------------------------------------------------------------|\n| --count            | count matching genomes only                                            |\n\n\nexample;\n```sh\n./sonar.py match --profile S:E484K --LINEAGE B.1.1.7 --db test.db\n\n# matching B.1.1.7 genomes in DB \'test.db\' that share an additional "Erik" mutation \n./sonar.py match --profile S:E484K --LINEAGE B.1.1.7 --db test.db\n\n# as before but matching genomes are counted only\n./sonar.py match --profile S:E484K --LINEAGE B.1.1.7 --count --db test.db\n\n# matching genomes in DB \'test.db\' sharing the "Nelly" mutation\n# and that were sampled in 2020\n./sonar.py match --profile S:N501Y  --DATE 2020-01-01:2020-12-31 --db test.db\n\n# matching genomes in DB \'mydb\' sharing the "Nelly" and the "Erik" mutation but not\n# belonging to the B.1.1.7 lineage\n./sonar.py match -profile S:N501Y S:E484K --LINEAGE ^B.1.1.7 --db test.db\n```\n\n**Export to CSV/TSV/VCF file**\n\ncovSonar can return results in different formats: `--format ["csv", "tsv", "vcf"]`\n\n```sh\n# example command\n./sonar.py match --profile S:N501Y S:E484K --LINEAGE ^B.1.1.7 --db test.db --format csv -o out.csv\n\n# in vcf format\n./sonar.py match -i S:N501Y S:E484K --lineage Q.1 --db test.db --format vcf -o out.vcf\n\n# example of --sample-file \n./sonar.py match --sample-file accessions.txt --db test.db --format vcf -o out.vcf\n```\n\n\n\n#### <u>Parent-Child relationship</u>\n\n> ⚠️ This function we only test on SARS-CoV-2\n\nIf we want to search all sublineages with a given lineage, covSonar offers `--with-sublineage PROP_COLUMN` (PROP_COLUMN  means the property name that we added to our database).\n\n```sh\n./sonar.py match --profle S:E484K --LINEAGE B.1.1.7 --with-sublineage LINEAGE --count --db test.db --debug\n```\nThis query will return results (\'B.1.1.7\', \'Q.4\', \'Q.5\', \'Q.3\', \'Q.6\', \'Q.1\', \'Q.7\', \'Q.2\', \'Q.8\').\n\nBy default, we use SARS-CoV-2 lineages for this search and the file name must be `lineage.all.tsv`.\n\n**lineage-update function for SARS-CoV-2 (COVID-19) ❗**\n\nRun `update-lineage-info` flag, it will download the latest version of lineages from https://github.com/cov-lineages/pango-designation/ and install it in `lib/lineage.all.tsv`\n\n```sh\n# example command\n./sonar.py update-lineage-info\n```\n\n### 3.5 Show infos about the used sonar system and database (info)\n\nDetailed infos about the used sonar system (e.g. version, reference,  number of imported genomes, unique sequences, available metadata).\n\n```sh\n# Show infos about the used sonar system and database \'test.db\'\n./sonar.py info --db test.db \n```\n\n### 3.6 Restore genome sequences from the database (restore)\nGenome sequences can be restored from the database based on their accessions.\nThe restored sequences are combined with their original FASTA header and  shown on the screen. The screen output can be redirected to a file easily by using `>`.\n\n```sh\n# Restore genome sequences linked to accessions \'mygenome1\' and \'mygenome2\' from the \n# database \'test.db\' and write these to a fasta file named \'restored.fasta\'\n./sonar.py restore --sample mygenome1 mygenome2 --db test.db > restored.fasta\n# as before, but consider all accessions from \'accessions.txt\' (the file has to\n# contain one accession per line) \n./sonar.py restore --sample-file accessions.txt --db test.db > restored.fasta\n```\n\n### 3.7 Database management (db-upgrade, optimize)\nSometimes you might need the `optimize` command to clean the [problems](https://www.sqlite.org/lang_vacuum.html) from database operation (e.g., unused data block or storage overhead ).\n```sh\n# Show infos about the used sonar system\n./sonar.py optimize  --db test.db \n```\n\nWhen the newest version of covSonar use an old database version, covSonar will return  the following error;\n```bash\nCompatibility error: the given database is not compatible with this version of sonar (Current database version: XXX; Supported database version: XXX)\nPlease run \'sonar.py  db-upgrade\' to upgrade database\n```\n\nWe provide our database upgrade assistant to solve the problem. \n```bash \n# RUN \n./sonar.py db-upgrade --db test.db\n\n# Output\nWarning: Backup db file before upgrading, Press Enter to continue...\n\n## after pressing the Enter key\nCurrent version: 3  Upgrade to: 4\nPerform the Upgrade: file: mydb.db\nDatabase now version: 4\nSuccess: Database upgrade was successfully completed\n\n```\n⚠️ Warning: Backup the db file before upgrade.\n\n\n## How to contribute 🏗️ \n\ncovSonar has been very carefully programmed and tested, but is still in an early stage of development. You can contribute to this project by reporting problems 🐛 or writing feature requests to the [issue section](https://github.com/rki-mf1/covsonar/issues) 👨\u200d💻\n\nYour feedback is very welcome 👨\u200d🔧!\n\n---------------------------------\n\n## Contact\n\nFor business inquiries or professional support requests 🍺 please contact [Dr. Stephan Fuchs](https://www.rki.de/SharedDocs/Personen/Mitarbeiter/F/Fuchs_Stephan.html)\n\n',
    'author': 'Stephan Fuchs',
    'author_email': 'FuchsS@rki.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rki-mf1/covsonar',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
